from typing import Any

import numpy as np
import pytest
from numpy.lib import NumpyVersion
from numpy.testing import assert_almost_equal, assert_array_equal

from pyoperators.utils.testing import assert_eq
from pysimulators.quantities import Quantity, UnitError

from .common import assert_equal_subclass


def assert_quantity_equal(q1: Any, q2: Any):
    assert isinstance(q1, Quantity)
    assert type(q1) is type(q2)
    assert_array_equal(q1.magnitude, q2.magnitude)
    assert q1._unit == q2._unit


def test_si():
    q = Quantity(1, 'km')
    assert_equal_subclass(q.SI, Quantity(1000, 'm'))


def test_tounit1():
    q = Quantity(1, 'm')
    assert_equal_subclass(q, q.tounit('m'))


def test_tounit2():
    q = Quantity(1, 'km')
    assert_equal_subclass(Quantity(1000, 'm'), q.tounit('m'))


def test_copy():
    q = Quantity(1, 'm')
    q2 = q.copy()
    q.inunit('m')
    assert_equal_subclass(q, q2)


def test_inunit():
    q = Quantity(1, 'km')
    q.inunit('m')
    assert_equal_subclass(Quantity(1000, 'm'), q)


@pytest.mark.parametrize('other', [Quantity(1.0), np.array(1.0), 1])
def test_add1(other):
    q = Quantity(1.0)
    assert_equal_subclass(q + other, Quantity(2))
    assert_equal_subclass(other + q, Quantity(2))


@pytest.mark.parametrize('other', [Quantity(1.0), np.array(1.0), 1])
def test_add2(other):
    q = Quantity(1.0, 'm')
    assert_equal_subclass(q + other, Quantity(2, 'm'))
    assert_equal_subclass(other + q, Quantity(2, 'm'))


def test_add3():
    q1 = Quantity(1.0)
    q2 = Quantity(1.0, 'km')
    assert_equal_subclass(q1 + q2, Quantity(2, 'km'))


def test_add4():
    q1 = Quantity(1.0, 'm')
    q2 = Quantity(1.0, 'km')
    assert_equal_subclass(q1 + q2, Quantity(1001, 'm'))


def test_add5():
    q1 = Quantity(1.0, 'km')
    q2 = Quantity(1.0, 'm')
    assert_equal_subclass(q1 + q2, Quantity(1.001, 'km'))


def test_add6():
    with pytest.raises(UnitError):
        Quantity(1, 'kloug') + Quantity(3.0, 'babar')


@pytest.mark.parametrize('other', [Quantity(1.0), np.array(1.0), 1])
def test_sub1(other):
    q = Quantity(1.0)
    assert_equal_subclass(q - other, Quantity(0.0))
    assert_equal_subclass(other - q, Quantity(0.0))


@pytest.mark.parametrize('other', [Quantity(1.0), np.array(1.0), 1])
def test_sub2(other):
    q = Quantity(1.0, 'm')
    assert_equal_subclass(q - other, Quantity(0.0, 'm'))
    assert_equal_subclass(other - q, Quantity(0.0, 'm'))


def test_sub3():
    q1 = Quantity(1.0)
    q2 = Quantity(1.0, 'km')
    assert_equal_subclass(q1 - q2, Quantity(0.0, 'km'))


def test_sub4():
    q1 = Quantity(1.0, 'm')
    q2 = Quantity(1.0, 'km')
    assert_equal_subclass(q1 - q2, Quantity(-999.0, 'm'))


def test_sub5():
    q1 = Quantity(1.0, 'km')
    q2 = Quantity(1.0, 'm')
    assert_equal_subclass(q1 - q2, Quantity(0.999, 'km'))


def test_sub6():
    with pytest.raises(UnitError):
        Quantity(1, 'kloug') - Quantity(3.0, 'babar')


def test_conversion_error():
    a = Quantity(1.0, 'kloug')
    with pytest.raises(UnitError):
        a.inunit('notakloug')
    with pytest.raises(UnitError):
        a.inunit('kloug^2')


def test_conversion_sr1():
    a = Quantity(1, 'MJy/sr').tounit('uJy/arcsec^2')
    assert_almost_equal(a, 23.5044305391)
    assert a.unit == 'uJy / arcsec^2'


def test_conversion_sr2():
    a = (Quantity(1, 'MJy/sr') / Quantity(1, 'uJy/arcsec^2')).SI
    assert_almost_equal(a, 23.5044305391)
    assert a.unit == ''


def test_array_ufunc():
    assert Quantity(10, 'm') <= Quantity(1, 'km')
    assert Quantity(10, 'm') < Quantity(1, 'km')
    assert Quantity(1, 'km') >= Quantity(10, 'm')
    assert Quantity(1, 'km') > Quantity(10, 'm')
    assert Quantity(1, 'km') != Quantity(1, 'm')
    assert Quantity(1, 'km') == Quantity(1000, 'm')
    assert np.maximum(Quantity(10, 'm'), Quantity(1, 'km')) == 1000
    assert np.minimum(Quantity(10, 'm'), Quantity(1, 'km')) == 10


@pytest.mark.parametrize(
    'func, args, keywords',
    [
        (np.amax, (), {}),
        (np.amin, (), {}),
        (np.around, (), {}),
        (np.atleast_1d, (), {}),
        (np.atleast_2d, (), {}),
        (np.atleast_3d, (), {}),
        (np.average, (), {}),
        (np.broadcast_to, ((3, 2, 2),), {}),
        (np.clip, (1, 2), {}),
        (np.copy, (), {}),
        (np.cumsum, (), {}),
        (getattr(np, 'cumulative_sum', 'cumulative_sum'), (), {'axis': 1}),
        (np.delete, ([0],), {}),
        (np.diag, (), {}),
        (np.diagflat, (), {}),
        (np.diagonal, (), {}),
        (np.diff, (), {}),
        (np.expand_dims, (), {'axis': 0}),
        (np.fix, (), {}),
        (np.flip, (), {}),
        (np.fliplr, (), {}),
        (np.flipud, (), {}),
        (np.imag, (), {}),
        (np.interp, ([1, 2], [1, 1.5]), {}),
        (np.max, (), {}),
        (np.mean, (), {}),
        (np.median, (), {}),
        (np.min, (), {}),
        (np.moveaxis, (0, 1), {}),
        (np.nan_to_num, (), {}),
        (np.nancumsum, (), {}),
        (np.nanmax, (), {}),
        (np.nanmean, (), {}),
        (np.nanmedian, (), {}),
        (np.nanmin, (), {}),
        (np.nanstd, (), {}),
        (np.nansum, (), {}),
        (np.pad, ([2, 3],), {}),
        (np.partition, (1,), {}),
        (np.permute_dims, (), {}),
        (np.ptp, (), {}),
        (np.ravel, (), {}),
        (np.real, (), {}),
        (np.real_if_close, (), {}),
        (np.repeat, (2,), {}),
        (np.reshape, ((4, 1),), {}),
        (np.resize, ((3, 3),), {}),
        (np.roll, (1,), {}),
        (np.rollaxis, (), {'axis': 1}),
        (np.rot90, (), {}),
        (np.round, (), {}),
        (np.sort, (), {}),
        (np.sort_complex, (), {}),
        (np.squeeze, (), {}),
        (np.std, (), {}),
        (np.sum, (), {}),
        (np.swapaxes, (0, 1), {}),
        (np.take, ([0, 0]), {}),
        (np.take_along_axis, (), {'axis': 0, 'indices': np.array([[0]])}),
        (np.tile, (2,), {}),
        (np.trace, (), {}),
        (np.transpose, (), {}),
        (np.trim_zeros, (), {}),
        (np.unique_values, (), {}),
        (np.unwrap, (), {}),
    ],
)
def test_function_default_conversion(func, args, keywords) -> None:
    if isinstance(func, str):
        pytest.skip(f'Function {func!r} unavailable for Numpy {np.__version__}')
    input = Quantity([[1, 2], [3, 4]], 'm')
    if NumpyVersion(np.__version__) < '2.2.0' and func is np.trim_zeros:
        input = input.ravel()

    actual_output = func(input, *args, **keywords)
    assert type(input) is Quantity
    assert_array_equal(
        actual_output.magnitude, func(input.view(np.ndarray), *args, **keywords)
    )
    assert actual_output._unit == {'m': 1.0}


@pytest.mark.parametrize(
    'func, inputs, expected_output',
    [
        (
            np.append,
            [Quantity([1, 2], 'm'), Quantity(1, 'km')],
            Quantity([1, 2, 1000], 'm'),
        ),
        (
            np.intersect1d,
            [Quantity([1, 2, 1000], 'm'), Quantity([1, 2], 'km')],
            Quantity([1000], 'm'),
        ),
        (
            np.setdiff1d,
            [Quantity([1, 2, 1000], 'm'), Quantity([1, 2], 'km')],
            Quantity([1, 2], 'm'),
        ),
        (
            np.setxor1d,
            [Quantity([1, 2, 1000], 'm'), Quantity([1, 2], 'km')],
            Quantity([1, 2, 2000], 'm'),
        ),
        (
            np.union1d,
            [Quantity([1, 2, 1000], 'm'), Quantity([1, 2], 'km')],
            Quantity([1, 2, 1000, 2000], 'm'),
        ),
    ],
)
def test_function_homogenize1(func, inputs, expected_output):
    actual_output = func(*inputs)
    assert_quantity_equal(actual_output, expected_output)


def test_function_fill_diagonal():
    array = Quantity.zeros((2, 2), 'm')
    diagonal = Quantity.ones(2, 'km')
    np.fill_diagonal(array, diagonal)
    assert_quantity_equal(array, Quantity([[1000, 0], [0, 1000]], 'm'))

    array = np.zeros((2, 2))
    diagonal = Quantity.ones(2, 'km')
    np.fill_diagonal(array, diagonal)
    assert_array_equal(array, np.array([[1, 0], [0, 1]]))


@pytest.mark.filterwarnings('ignore:`row_stack` alias is deprecated')
@pytest.mark.parametrize(
    'func, q1, q2, expected_output',
    [
        (
            np.column_stack,
            Quantity([1, 2], 'm'),
            Quantity([2, 3], 'km'),
            Quantity([[1, 2000], [2, 3000]], 'm'),
        ),
        (
            np.concat,
            Quantity([1, 2], 'm'),
            Quantity([3, 4], 'km'),
            Quantity([1, 2, 3000, 4000], 'm'),
        ),
        (
            np.concatenate,
            Quantity([1, 2], 'm'),
            Quantity([3, 4], 'km'),
            Quantity([1, 2, 3000, 4000], 'm'),
        ),
        (
            np.dstack,
            Quantity([[1], [2], [3]], 'm'),
            Quantity([[1], [2], [3]], 'km'),
            Quantity([[[1, 1000]], [[2, 2000]], [[3, 3000]]], 'm'),
        ),
        (
            np.hstack,
            Quantity([[1], [2]], 'm'),
            Quantity([[2], [3]], 'km'),
            Quantity([[1, 2000], [2, 3000]], 'm'),
        ),
        (
            np.row_stack,
            Quantity([1, 2], 'm'),
            Quantity([2, 3], 'km'),
            Quantity([[1, 2], [2000, 3000]], 'm'),
        ),
        (
            np.stack,
            Quantity([1], 'm'),
            Quantity([1], 'km'),
            Quantity([[1], [1000]], 'm'),
        ),
        (
            np.vstack,
            Quantity([1, 2], 'm'),
            Quantity([2, 3], 'km'),
            Quantity([[1, 2], [2000, 3000]], 'm'),
        ),
    ],
)
def test_function_homogenize2(func, q1, q2, expected_output):
    actual_output = func([q1, q2])
    assert_quantity_equal(actual_output, expected_output)


@pytest.mark.parametrize(
    'func, array, args',
    [
        (np.array_split, Quantity([[1, 2], [3, 4]], 'm'), (2,)),
        (np.dsplit, Quantity([[[1, 2], [3, 4]]], 'm'), (2,)),
        (np.hsplit, Quantity([[1, 2], [3, 4]], 'm'), (2,)),
        (np.split, Quantity([[1, 2], [3, 4]], 'm'), (2,)),
        (np.vsplit, Quantity([[1, 2], [3, 4]], 'm'), (2,)),
        (getattr(np, 'unstack', 'unstack'), Quantity([[1, 2], [3, 4]], 'm'), ()),
    ],
)
def test_function_output_sequence(func, array, args):
    if isinstance(func, str):
        pytest.skip(f'Function {func!r} unavailable for Numpy {np.__version__}')
    actual_outputs = func(array, *args)
    expected_outputs = func(array.view(np.ndarray), *args)
    assert isinstance(actual_outputs, (list, tuple))
    assert len(actual_outputs) == len(expected_outputs)
    for actual_output, expected_output in zip(actual_outputs, expected_outputs):
        assert_quantity_equal(actual_output, Quantity(expected_output, 'm'))


@pytest.mark.parametrize(
    'func, keywords',
    [
        (np.convolve, {}),
        (np.correlate, {}),
        (np.cov, {}),
        (np.cross, {}),
        (np.dot, {}),
        (np.inner, {}),
        (np.kron, {}),
        (np.outer, {}),
        (np.tensordot, {'axes': 0}),
        (np.vdot, {}),
    ],
)
def test_function_action_prod(func, keywords):
    array = Quantity([1, 2, 3], 'm')
    actual_output = func(array, array, **keywords)
    expected_output = func(array.view(np.ndarray), array.view(np.ndarray), **keywords)
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm^2'))


@pytest.mark.parametrize(
    'func', [np.min, np.max, np.mean, np.ptp, np.sum, np.std, np.var]
)
def test_function_with_axis(func):
    a = Quantity([[1.0, 2, 3], [4, 5, 6]], unit='Jy')
    actual_output = func(a, axis=0)
    expected_unit = 'Jy^2' if func is np.var else 'Jy'
    expected_output = Quantity(func(a.view(np.ndarray), axis=0), expected_unit)
    assert_quantity_equal(actual_output, expected_output)


def test_function_broadcast_arrays():
    q1 = Quantity([1, 2], 'm')
    q2 = Quantity([[1], [2], [3]], 's')
    actual_outputs = np.broadcast_arrays(q1, q2)
    expected_outputs = np.broadcast_arrays(q1.view(np.ndarray), q2.view(np.ndarray))
    assert isinstance(actual_outputs, (list, tuple))
    assert len(actual_outputs) == 2
    assert_quantity_equal(actual_outputs[0], Quantity(expected_outputs[0], 'm'))
    assert_quantity_equal(actual_outputs[1], Quantity(expected_outputs[1], 's'))


def test_function_choose():
    choices = [
        Quantity([1, 2, 3], 'm'),
        Quantity([4, 5, 6], 'km'),
        Quantity([7, 8, 9], 'm'),
    ]
    indices = np.array([0, 1, 2])
    actual_output = np.choose(indices, choices)
    # Should homogenize to 'm' (first non-empty unit) and convert km to m
    homogenized_choices = [
        choices[0].view(np.ndarray),
        choices[1].tounit('m').view(np.ndarray),
        choices[2].view(np.ndarray),
    ]
    expected_output = np.choose(indices, homogenized_choices)
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm'))


@pytest.mark.parametrize('func', [np.compress, np.extract])
def test_function_condition_second(func):
    condition = [True, False, True]
    array = Quantity([1, 2, 3], 'm')
    actual_output = func(condition, array)
    expected_output = func(condition, array.view(np.ndarray))
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm'))


def test_function_einsum():
    a = Quantity([1, 2], 'm')
    b = Quantity([3, 4], 's')
    actual_output = np.einsum('i,i->', a, b)
    expected_output = np.einsum('i,i->', a.view(np.ndarray), b.view(np.ndarray))
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm s'))


def test_function_insert():
    array = Quantity([1, 2, 3], 'm')
    values = Quantity([100, 200], 'km')
    actual_output = np.insert(array, 1, values)
    expected_output = np.insert(
        array.view(np.ndarray), 1, values.tounit('m').view(np.ndarray)
    )
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm'))


@pytest.mark.parametrize('func', [np.var, np.nanvar])
def test_function_action_square(func):
    array = Quantity([1, 2, 3], 'm')
    actual_output = func(array)
    expected_output = func(array.view(np.ndarray))
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm^2'))


def test_function_select():
    condlist = [np.array([True, False, False]), np.array([False, True, False])]
    choicelist = [Quantity([1, 2, 3], 'm'), Quantity([4, 5, 6], 'km')]
    actual_output = np.select(condlist, choicelist, default=0)
    expected_output = np.select(
        condlist,
        [choicelist[0].view(np.ndarray), choicelist[1].tounit('m').view(np.ndarray)],
        default=0,
    )
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm'))


def test_function_unique():
    array = Quantity([1, 2, 1, 3, 2], 'm')
    # Test with return_index=False (returns single array)
    actual_output = np.unique(array)
    expected_output = np.unique(array.view(np.ndarray))
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm'))

    # Test with return_index=True (returns tuple)
    actual_output, actual_indices = np.unique(array, return_index=True)
    expected_output, expected_indices = np.unique(
        array.view(np.ndarray), return_index=True
    )
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm'))
    assert_array_equal(actual_indices, expected_indices)


def test_function_unique_all():
    array = Quantity([1, 2, 1, 3, 2], 'm')
    result = np.unique_all(array)

    assert isinstance(result, np.lib._arraysetops_impl.UniqueAllResult)
    assert type(result.values) is Quantity
    assert type(result.indices) is np.ndarray
    assert type(result.inverse_indices) is np.ndarray
    assert type(result.counts) is np.ndarray

    expected = np.unique_all(array.view(np.ndarray))
    assert_quantity_equal(result.values, Quantity(expected.values, 'm'))
    assert_array_equal(result.indices, expected.indices)
    assert_array_equal(result.inverse_indices, expected.inverse_indices)
    assert_array_equal(result.counts, expected.counts)


def test_function_where():
    condition = np.array([True, False, True])
    x = Quantity([1, 2, 3], 'm')
    y = Quantity([4, 5, 6], 'km')
    actual_output = np.where(condition, x, y)
    expected_output = np.where(
        condition, x.view(np.ndarray), y.tounit('m').view(np.ndarray)
    )
    assert_quantity_equal(actual_output, Quantity(expected_output, 'm'))


@pytest.mark.parametrize(
    'value, expected_type',
    (
        [
            (Quantity(1), float),
            (Quantity(1, dtype='float32'), np.float32),
            (Quantity(1.0), float),
            (Quantity(complex(1, 0)), np.complex128),
            (Quantity(1.0, dtype=np.complex64), np.complex64),
            (Quantity(1.0, dtype=np.complex128), np.complex128),
            (Quantity(np.array(complex(1, 0))), complex),
            (Quantity(np.array(np.complex64(1.0))), np.complex64),
            (Quantity(np.array(np.complex128(1.0))), complex),
        ]
        + [
            (Quantity(1.0, dtype=np.complex256), np.complex256),
        ]
        if hasattr(np, 'complex256')
        else []
    ),
)
def test_dtype(value, expected_type):
    assert value.dtype == expected_type


def test_derived_units1():
    du = {
        'detector[rightward]': Quantity([1, 1 / 10.0], 'm^2'),
        'time[leftward]': Quantity([1, 1 / 2, 1 / 3, 1 / 4], 's'),
    }
    a = Quantity([[1, 2, 3, 4], [10, 20, 30, 40]], 'detector', derived_units=du)
    assert_equal_subclass(a.SI, Quantity([[1, 2, 3, 4], [1, 2, 3, 4]], 'm^2', du))

    a = Quantity([[1, 2, 3, 4], [10, 20, 30, 40]], 'time', derived_units=du)
    assert_equal_subclass(a.SI, Quantity([[1, 1, 1, 1], [10, 10, 10, 10]], 's', du))

    a = Quantity(1, 'detector C', du)
    assert a.SI.shape == (2,)

    a = Quantity(np.ones((1, 10)), 'detector', derived_units=du)
    assert a.SI.shape == (2, 10)


def test_derived_units2():
    a = Quantity(4.0, 'Jy/detector', {'detector': Quantity(2, 'arcsec^2')})
    a.inunit(a.unit + ' / arcsec^2 * detector')
    assert_quantity_equal(a, Quantity(2, 'Jy / arcsec^2'))


def test_derived_units3():
    a = Quantity(
        2.0,
        'brou',
        {
            'brou': Quantity(2.0, 'bra'),
            'bra': Quantity(2.0, 'bri'),
            'bri': Quantity(2.0, 'bro'),
            'bro': Quantity(2, 'bru'),
            'bru': Quantity(2.0, 'stop'),
        },
    )
    b = a.tounit('bra')
    assert b.magnitude == 4
    b = a.tounit('bri')
    assert b.magnitude == 8
    b = a.tounit('bro')
    assert b.magnitude == 16
    b = a.tounit('bru')
    assert b.magnitude == 32
    b = a.tounit('stop')
    assert b.magnitude == 64
    b = a.SI
    assert b.magnitude == 64
    assert b.unit == 'stop'


_ = slice(None)
DERIVED_UNITS_KEYS = [
    1,
    (1,),
    (1, 2),
    (1, 2, 3),
    (1, 2, 3, 4),
    _,
    (_,),
    (_, 1),
    (_, 1, 2),
    (_, 1, 2, 3),
    (1, _),
    (1, _, 2),
    (1, _, 2, 3),
    (1, 2, _),
    (1, 2, _, 3),
    (1, 2, 3, _),
    (_, _),
    (_, _, 1),
    (_, 1, _),
    (1, _, _),
    (_, _, 1, 2),
    (_, 1, _, 2),
    (_, 1, 2, _),
    (1, _, _, 2),
    (1, _, 2, _),
    (1, 2, _, _),
    (_, _, _),
    (_, _, _, 1),
    (_, _, 1, _),
    (_, 1, _, _),
    (1, _, _, _),
    (_, _, _, _),
    Ellipsis,
]


@pytest.mark.parametrize('key', DERIVED_UNITS_KEYS)
def test_derived_units_leftward(key):
    l = np.arange(3 * 4 * 5).reshape((3, 4, 5))
    a_leftward = Quantity.ones((2, 3, 4, 5), 'r', {'r[leftward]': Quantity(l)})

    b = a_leftward[key]
    if np.isscalar(b):
        assert b.dtype == np.float64
        return
    if key is Ellipsis:
        assert_eq(
            b.derived_units['r[leftward]'], a_leftward.derived_units['r[leftward]']
        )
    if key in ((1, 2, 3, 4), (_, 1, 2, 3)):
        assert 'r' in b.derived_units
        if key == (1, 2, 3, 4):
            assert b.derived_units['r'] == Quantity(59)
        else:
            assert b.derived_units['r'] == Quantity(33)
    else:
        assert 'r' not in b.derived_units
        if not isinstance(key, tuple):
            key = (key,)
        key += (4 - len(key)) * (slice(None),)
        assert_eq(
            b.derived_units['r[leftward]'],
            a_leftward.derived_units['r[leftward]'][key[-3:]],
        )


@pytest.mark.parametrize('key', DERIVED_UNITS_KEYS)
def test_derived_units_rightward(key):
    r = np.arange(2 * 3 * 4).reshape((2, 3, 4))
    a_rightward = Quantity.ones((2, 3, 4, 5), 'r', {'r[rightward]': Quantity(r)})

    b = a_rightward[key]
    if np.isscalar(b):
        assert b.dtype == np.float64
        return
    if key is Ellipsis:
        assert_eq(
            b.derived_units['r[rightward]'], a_rightward.derived_units['r[rightward]']
        )
    if key in ((1, 2, 3), (1, 2, 3, 4), (1, 2, 3, _)):
        assert 'r' in b.derived_units
        assert b.derived_units['r'] == Quantity(23)
    else:
        assert 'r' not in b.derived_units
        if not isinstance(key, tuple):
            key = (key,)
        assert_eq(
            b.derived_units['r[rightward]'],
            a_rightward.derived_units['r[rightward]'][key[:3]],
        )


def test_pixels():
    actual = Quantity(1, 'pixel/sr/pixel_reference').SI
    expected = Quantity(1, 'sr^-1')
    assert_equal_subclass(actual, expected)



import numpy as np
import pytest
import scipy
import scipy.constants
from astropy.coordinates.angles import Angle
from astropy.time import Time
from numpy.testing import assert_allclose, assert_equal

from pyoperators import (
    BlockDiagonalOperator,
    Cartesian2SphericalOperator,
    CompositionOperator,
    IdentityOperator,
    MultiplicationOperator,
    Operator,
    Spherical2CartesianOperator,
    flags,
)
from pyoperators.utils import isscalarlike, product
from pyoperators.utils.testing import assert_same
from pysimulators.operators import (
    BlackBodyOperator,
    CartesianEquatorial2GalacticOperator,
    CartesianEquatorial2HorizontalOperator,
    CartesianGalactic2EquatorialOperator,
    CartesianHorizontal2EquatorialOperator,
    ConvolutionTruncatedExponentialOperator,
    PowerLawOperator,
    RollOperator,
    SphericalEquatorial2GalacticOperator,
    SphericalEquatorial2HorizontalOperator,
    SphericalGalactic2EquatorialOperator,
    SphericalHorizontal2EquatorialOperator,
    block_diagonal,
)


@block_diagonal('value', 'key', axisin=0)
@flags.square
class PartitionChunkOp(Operator):
    def __init__(self, arg1, value, arg3, key=None, **keywords):
        Operator.__init__(self, **keywords)
        self.arg1 = arg1
        self.value = value
        self.arg3 = arg3
        self.key = key

    def direct(self, input, output):
        output[...] = self.value * input


class PartitionChunkSubOp(PartitionChunkOp):
    pass


@pytest.mark.parametrize('cls', [PartitionChunkOp, PartitionChunkSubOp])
@pytest.mark.parametrize('partitionin', [None, 2, (2,), (4, 2)])
@pytest.mark.parametrize('value', [2.0, (2.0,), (2.0, 3)])
@pytest.mark.parametrize('key', [0.0, (0.0,), (0.0, 1.0)])
def test_partitioning_chunk(cls, partitionin, value, key):
    def assert_partioning_chunk(cls, op, value, key):
        assert op.__class__ is cls
        assert op.arg1 is arg1
        assert op.value is value
        assert hasattr(op, 'arg3')
        assert op.arg3 is arg3
        assert op.key is key
        input = np.ones(1)
        output = op(input)
        assert_equal(output, value)

    arg1 = [1, 2, 3, 4, 5]
    arg3 = ['a', 'b', 'c', 'd']

    n1 = 1 if isscalarlike(value) else len(value)
    n2 = 1 if isscalarlike(key) else len(key)
    nn = (
        max(n1, n2)
        if partitionin is None
        else 1 if isscalarlike(partitionin) else len(partitionin)
    )
    if n1 != n2 and not isscalarlike(value) and not isscalarlike(key):
        # the partitioned arguments do not have the same length
        with pytest.raises(
            ValueError,
            match='The partition variables do not have the same number of elements',
        ):
            cls(arg1, value, arg3, key=key, partitionin=partitionin)
        return
    if nn != max(n1, n2) and (not isscalarlike(value) or not isscalarlike(key)):
        # the partition is incompatible with the partitioned arguments
        with pytest.raises(
            ValueError,
            match='The specified partitioning is incompatible with the number of elements in the partition variables.',
        ):
            cls(arg1, value, arg3, key=key, partitionin=partitionin)
        return

    op = cls(arg1, value, arg3, key=key, partitionin=partitionin)
    if nn == 1:
        value = value if isscalarlike(value) else value[0]
        key = key if isscalarlike(key) else key[0]
        assert_partioning_chunk(cls, op, value, key)
        return

    assert op.__class__ is BlockDiagonalOperator
    assert len(op.operands) == nn
    if partitionin is None:
        assert op.partitionin == nn * (None,)
        assert op.partitionout == nn * (None,)
        return
    value = nn * [value] if isscalarlike(value) else value
    key = nn * [key] if isscalarlike(key) else key
    for op_, v_, k_ in zip(op.operands, value, key):
        assert_partioning_chunk(cls, op_, v_, k_)
    expected = np.hstack([n_ * [v_] for n_, v_ in zip(partitionin, value)])
    input = np.ones(np.sum(partitionin))
    output = op(input)
    assert_equal(output, expected)


@block_diagonal('value', 'key', new_axisin=0)
@flags.square
class PartitionStackOp(Operator):
    def __init__(self, arg1, value, arg3, key=None, **keywords):
        Operator.__init__(self, **keywords)
        self.arg1 = arg1
        self.value = value
        self.arg3 = arg3
        self.key = key

    def direct(self, input, output):
        output[...] = self.value * input


class PartitionStackSubOp(PartitionStackOp):
    pass


@pytest.mark.parametrize('cls', [PartitionStackOp, PartitionStackSubOp])
@pytest.mark.parametrize('value', [2.0, (2.0,), (2.0, 3)])
@pytest.mark.parametrize('key', [0.0, (0.0,), (0.0, 1.0)])
def test_partitioning_stack(cls, value, key):
    def assert_partition_stack(cls, op, v, k):
        assert op.__class__ is cls
        assert op.arg1 is arg1
        assert op.value is v
        assert hasattr(op, 'arg3')
        assert op.arg3 is arg3
        assert op.key is k
        input = np.ones(1)
        output = op(input)
        assert_equal(output, v)

    arg1 = [1, 2, 3, 4, 5]
    arg3 = ['a', 'b', 'c', 'd']

    n1 = 1 if isscalarlike(value) else len(value)
    n2 = 1 if isscalarlike(key) else len(key)
    nn = max(n1, n2)
    if not isscalarlike(value) and not isscalarlike(key) and n1 != n2:
        # the partitioned arguments do not have the same length
        with pytest.raises(ValueError):
            cls(arg1, value, arg3, key=key)
        return

    op = cls(arg1, value, arg3, key=key)
    if nn == 1:
        value = value if isscalarlike(value) else value[0]
        key = key if isscalarlike(key) else key[0]
        assert_partition_stack(cls, op, value, key)
        return

    assert op.__class__ is BlockDiagonalOperator
    assert len(op.operands) == nn
    value = nn * [value] if isscalarlike(value) else value
    key = nn * [key] if isscalarlike(key) else key
    for op_, v_, k_ in zip(op.operands, value, key):
        assert_partition_stack(cls, op_, v_, k_)
    input = np.ones(nn)
    output = op(input)
    assert_equal(output, value)


def ref_blackbody(w, T):
    c = scipy.constants.c
    h = scipy.constants.h
    k = scipy.constants.k
    nu = c / w
    return 2 * h * nu**3 / c**2 / (np.exp(h * nu / (k * T)) - 1)


def test_blackbody_scalar():

    w = np.arange(90.0, 111) * 1e-6
    T = 15.0
    flux = ref_blackbody(w, T) / ref_blackbody(100e-6, T)
    ops = [BlackBodyOperator(T, wavelength=wave, wavelength0=100e-6) for wave in w]
    flux2 = [op(1.0) for op in ops]
    assert_same(flux, flux2)


def test_blackbody_array():
    w, T = np.ogrid[90:111, 15:20]
    w = w * 1.0e-6
    flux = ref_blackbody(w, T) / ref_blackbody(100e-6, T)
    ops = [
        BlackBodyOperator(T.squeeze(), wavelength=wave[0], wavelength0=100e-6)
        for wave in w
    ]
    flux2 = np.array([op(np.ones(T.size)) for op in ops])
    assert_same(flux, flux2)


def test_convolution_truncated_exponential():
    tau = (2, 3, 0)
    r = ConvolutionTruncatedExponentialOperator(tau, shapein=(3, 10))
    a = np.ones((3, 10))
    b = r(a)
    assert np.allclose(a, b)

    a[:, 1:] = 0
    b = r(a)
    assert_same(b[:2, :9], [np.exp(-np.arange(9) / _) for _ in tau[:2]])
    assert_same(b[2], a[2])
    assert_same(r.T.todense(), r.todense().T)
    assert_same(r.T.todense(inplace=True), r.todense(inplace=True).T)


@pytest.mark.parametrize('tau', [0, [0], (0,), np.array(0), np.array([0, 0])])
def test_convolution_truncated_exponential_morphing(tau):
    op = ConvolutionTruncatedExponentialOperator(tau)
    assert type(op) is IdentityOperator


@pytest.mark.parametrize('tau', [[0, 1], (0, 1), np.array([0, 1])])
def test_convolution_truncated_exponential_not_morphing(tau):
    op = ConvolutionTruncatedExponentialOperator(tau)
    assert type(op) is ConvolutionTruncatedExponentialOperator


@pytest.mark.parametrize('alpha', [-1, np.arange(11) - 5])
def test_power_law(alpha):
    c = scipy.constants.c
    nu = c / 120e-6
    nu0 = c / 100e-6
    values = np.arange(11)

    expected = (nu / nu0) ** alpha * values
    op = PowerLawOperator(alpha, nu, nu0)
    assert_same(op(values), expected)


@pytest.mark.parametrize('cls', [CompositionOperator, MultiplicationOperator])
def test_power_law_absorb_scalar(cls):
    op = cls([2, PowerLawOperator(1, 2.0, 3.0)])
    assert isinstance(op, PowerLawOperator)


@pytest.mark.parametrize('n', range(4))
@pytest.mark.parametrize(
    'axis',
    [
        (0,),
        (1,),
        (2,),
        (3,),
        (0, 1),
        (0, 2),
        (0, 3),
        (1, 2),
        (1, 3),
        (2, 3),
        (0, 1, 2),
        (0, 1, 3),
        (0, 2, 3),
        (1, 2, 3),
        (1, 2, 3),
        (0, 1, 2, 3),
    ],
)
def test_roll(n, axis):
    shape = np.arange(2, 6)
    v = np.arange(2 * 3 * 4 * 5).reshape(shape)
    expected = v.copy()
    for a in axis:
        expected = np.roll(expected, n, a)
    result = RollOperator(axis=axis, n=n)(v)
    assert_equal(result, expected)


@pytest.mark.parametrize(
    'op',
    [CartesianEquatorial2GalacticOperator(), CartesianGalactic2EquatorialOperator()],
)
@pytest.mark.parametrize('shape', [3, (2, 3), (4, 2, 3)])
def test_equ2gal(op, shape):
    assert op.I is op.T
    assert_same(op.todense(shapein=3), op.data)

    vec = np.arange(product(shape)).reshape(shape)
    vec_ = vec.reshape((-1, 3))
    expected = np.empty(shape)
    expected_ = expected.reshape((-1, 3))
    for i in range(expected.size // 3):
        expected_[i] = op(vec_[i])
    actual = op(vec)
    assert_same(actual, expected)


def test_equ2gal_rules():
    equ2gal = CartesianEquatorial2GalacticOperator()
    gal2equ = CartesianGalactic2EquatorialOperator()

    assert_same(gal2equ.todense(shapein=3), equ2gal.data.T)
    assert_same(equ2gal.todense(shapein=3), gal2equ.data.T)

    assert isinstance(equ2gal(gal2equ), IdentityOperator)
    assert isinstance(gal2equ(equ2gal), IdentityOperator)


def test_equ2hor():
    lat = 52
    lon = -64
    date = Time('1980-04-22 14:36:51.67', scale='ut1')
    E2h = CartesianEquatorial2HorizontalOperator
    gst = E2h._jd2gst(date.jd)
    lst = E2h._gst2lst(gst, lon)

    # Duffett-Smith §21
    assert_allclose(gst, 4.668119)
    assert_allclose(lst, 0.401453, rtol=1e-6)

    ra = Angle(lst - 5.862222, unit='hour').radian
    dec = Angle('23d13m10s').radian
    s2c = Spherical2CartesianOperator('azimuth,elevation')
    op = CartesianEquatorial2HorizontalOperator('NE', date, lat, lon)
    incoords = s2c([ra, dec])
    outcoords = op(incoords)
    assert_same(op.I(outcoords), incoords)
    az, el = np.degrees(s2c.I(outcoords))

    # Duffett-Smith §25
    assert_allclose(az % 360, 283.271027)
    assert_allclose(el, 19.334345, rtol=1e-6)


@pytest.mark.parametrize(
    'cls_sph, cls_car',
    [
        (SphericalEquatorial2GalacticOperator, CartesianEquatorial2GalacticOperator),
        (
            SphericalEquatorial2HorizontalOperator,
            CartesianEquatorial2HorizontalOperator,
        ),
        (SphericalGalactic2EquatorialOperator, CartesianGalactic2EquatorialOperator),
        (
            SphericalHorizontal2EquatorialOperator,
            CartesianHorizontal2EquatorialOperator,
        ),
    ],
)
@pytest.mark.parametrize(
    'conventionin, dirs, shape',
    [
        ('zenith,azimuth', (0, 0), ()),
        ('zenith,azimuth', (20, 0), ()),
        ('zenith,azimuth', (130, 0), ()),
        ('zenith,azimuth', (10, 20), ()),
        ('zenith,azimuth', (20, 190), ()),
        ('zenith,azimuth', [(0, 0), (20, 0), (130, 0), (10, 20), (20, 130)], (5,)),
        ('zenith,azimuth', [[(0, 0), (20, 200), (130, 300)]], (1, 3)),
        ('azimuth,zenith', (0, 0), ()),
        ('azimuth,zenith', (0, 20), ()),
        ('azimuth,zenith', (0, 130), ()),
        ('azimuth,zenith', (20, 10), ()),
        ('azimuth,zenith', (190, 20), ()),
        ('azimuth,zenith', [(0, 0), (0, 20), (0, 130), (20, 10), (130, 20)], (5,)),
        ('azimuth,zenith', [[(0, 0), (200, 20), (300, 130)]], (1, 3)),
        ('elevation,azimuth', (90, 0), ()),
        ('elevation,azimuth', (70, 0), ()),
        ('elevation,azimuth', (-40, 0), ()),
        ('elevation,azimuth', (80, 20), ()),
        ('elevation,azimuth', (70, 190), ()),
        ('elevation,azimuth', [(90, 0), (70, 0), (-40, 0), (80, 20), (70, 130)], (5,)),
        ('elevation,azimuth', [[(90, 0), (70, 200), (-40, 300)]], (1, 3)),
        ('azimuth,elevation', (0, 90), ()),
        ('azimuth,elevation', (0, 70), ()),
        ('azimuth,elevation', (0, -40), ()),
        ('azimuth,elevation', (20, 80), ()),
        ('azimuth,elevation', (190, 70), ()),
        ('azimuth,elevation', [(0, 90), (0, 70), (0, -40), (20, 80), (130, 70)], (5,)),
        ('azimuth,elevation', [[(0, 90), (200, 70), (300, -40)]], (1, 3)),
    ],
)
@pytest.mark.parametrize(
    'conventionout',
    [
        'zenith,azimuth',
        'azimuth,zenith',
        'elevation,azimuth',
        'azimuth,elevation',
    ],
)
@pytest.mark.parametrize('degrees', [False, True])
def test_spherical(cls_sph, cls_car, conventionin, dirs, shape, conventionout, degrees):
    if 'Horizontal' in str(cls_sph):
        args = ('NE', Time('1980-04-22 14:36:51.67', scale='ut1'), 100.1, -80)
    else:
        args = ()
    op_sph = cls_sph(
        *args, conventionin=conventionin, conventionout=conventionout, degrees=degrees
    )
    actual = op_sph(dirs)
    assert_equal(actual.shape, shape + (2,))
    if degrees:
        dirs = np.radians(dirs)
    expected = Cartesian2SphericalOperator(conventionout)(
        cls_car(*args)(Spherical2CartesianOperator(conventionin)(dirs))
    )
    if degrees:
        np.degrees(expected, expected)
    assert_same(actual, expected)



import numpy as np
import pytest
from numpy.testing import assert_allclose

from pyoperators.utils.testing import assert_same
from pysimulators.noises import (
    _fold_psd,
    _gaussian_psd_1f,
    _gaussian_sample,
    _interp,
    _unfold_psd,
)


def test_fold():
    p = [1, 1, 1, 1]
    p_folded = _fold_psd(p)
    assert_same(p_folded, [1, 2, 1])


def test_unfold():
    p = [1, 2, 1]
    p_unfolded = _unfold_psd(p)
    assert_same(p_unfolded, [1, 1, 1, 1])


@pytest.mark.parametrize(
    'z',
    [
        [-1, -0.5, 0, 0.5, 1, 1.5, 2, 2.5, 3, 3.5, 4, 4.5, 7.5, 8],
        [-1, -0.5],
        [2.25, 2.5],
        [8, 10],
    ],
)
def test_interp(z):
    x = [1, 2, 3, 5, 7]
    y = [[1, 2, 3, 5, 7], [0, 1, 2, 4, 6]]

    out1 = _interp(z, x, y[0])
    assert_same(z, out1)
    out2 = _interp(z, x, y[1])
    out = _interp(z, x, y)
    assert_same(out, np.vstack([out1, out2]))


def test_gaussian():
    fsamp = 5
    nsamples = 2**10
    white = [0.3, 0.5]
    fknee = 0.0
    alpha = 1

    def func(twosided):
        np.random.seed(0)
        p = _gaussian_psd_1f(nsamples, fsamp, white, fknee, alpha, twosided=twosided)
        t = _gaussian_sample(nsamples, fsamp, p, twosided=twosided)
        return p, t

    p1, t1 = func(False)
    p2, t2 = func(True)

    bandwidth = fsamp / nsamples
    assert_same(
        np.sum(p1, axis=-1) * bandwidth, np.sum(p2, axis=-1) * bandwidth, rtol=1000
    )

    assert_allclose(np.sqrt(np.sum(p1, axis=-1) * bandwidth), white, rtol=1e-3)
    assert_same(t1, t2)



import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from pyoperators import CompositionOperator, IdentityOperator
from pyoperators.utils.testing import assert_same
from pysimulators.interfaces.healpy import (
    Cartesian2HealpixOperator,
    Healpix2CartesianOperator,
    Healpix2SphericalOperator,
    HealpixConvolutionGaussianOperator,
    HealpixLaplacianOperator,
    Spherical2HealpixOperator,
)

hp = pytest.importorskip('healpy')

NSIDE = 512


@pytest.mark.parametrize(
    'vec, shape',
    [
        ((1, 0, 0), ()),
        ((0, 1, 0), ()),
        ((0, 0, 1), ()),
        (((1, 0, 0), (0, 1, 0), (0, 0, 1)), (3,)),
        ([(1, 0, 0), (0, 1, 0), (0, 0, 1)], (3,)),
        ((((1, 0, 0), (0, 1, 0)),), (1, 2)),
        ([[(1, 0, 0), (0, 1, 0)]], (1, 2)),
    ],
)
@pytest.mark.parametrize('nest', [False, True])
def test_cartesian_healpix(vec, shape, nest):

    c2h = Cartesian2HealpixOperator(NSIDE, nest=nest)
    h2c = Healpix2CartesianOperator(NSIDE, nest=nest)
    a = h2c(c2h(vec))
    assert_equal(a.shape, shape + (3,))
    assert_allclose(a, vec, atol=1e-1)


def test_cartesian_healpix_rules():
    op = Cartesian2HealpixOperator(NSIDE)
    assert type(op.I) is Healpix2CartesianOperator
    assert op.I.nside == op.nside
    assert (
        type(Cartesian2HealpixOperator(NSIDE)(Healpix2CartesianOperator(2 * NSIDE)))
        is CompositionOperator
    )

    assert (
        type(Cartesian2HealpixOperator(NSIDE)(Healpix2CartesianOperator(NSIDE)))
        is IdentityOperator
    )


@pytest.mark.parametrize('shapein', [(), (1,), (2,), (3,)])
@pytest.mark.parametrize('shapeout', [(), (1,), (2,), (3,)])
def test_cartesian_healpix_invalid_shape(shapein, shapeout):
    op = Cartesian2HealpixOperator(NSIDE)
    input = np.zeros(shapein)
    output = np.zeros(shapeout)

    if shapein == (3,) and shapeout == ():
        op(input, output)
        return

    with pytest.raises(ValueError):
        op(input, output)


@pytest.mark.parametrize('nest', [False, True])
@pytest.mark.parametrize(
    'vec, shape',
    [
        (0, ()),
        (100, ()),
        (12 * NSIDE**2 - 1, ()),
        ([0, 100, 12 * NSIDE**2 - 1], (3,)),
        ([[0, 100]], (1, 2)),
    ],
)
def test_healpix_cartesian(nest, vec, shape):
    h2c = Healpix2CartesianOperator(NSIDE, nest=nest)
    c2h = Cartesian2HealpixOperator(NSIDE, nest=nest)
    a = c2h(h2c(vec))
    assert a.shape == shape
    assert_allclose(a, vec)


def test_healpix_cartesian_rules():
    op = Healpix2CartesianOperator(NSIDE)
    assert type(op.I) is Cartesian2HealpixOperator
    assert op.I.nside == op.nside
    assert (
        type(Healpix2CartesianOperator(NSIDE)(Cartesian2HealpixOperator(2 * NSIDE)))
        is CompositionOperator
    )

    assert (
        type(Healpix2CartesianOperator(NSIDE)(Cartesian2HealpixOperator(NSIDE)))
        is IdentityOperator
    )


@pytest.mark.parametrize('shapein', [(), (1,), (2,), (3,)])
@pytest.mark.parametrize('shapeout', [(), (1,), (2,), (3,)])
def test_healpix_cartesian_invalid_shape(shapein, shapeout):
    op = Healpix2CartesianOperator(NSIDE)
    input = np.zeros(shapein)
    output = np.zeros(shapeout)

    if shapein == () and shapeout == (3,):
        op(input, output)
        return

    with pytest.raises(ValueError):
        op(input, output)


@pytest.mark.parametrize('pixel', [-1, 12 * NSIDE**2])
def test_healpix_cartesian_invalid_pixel(pixel):
    op = Healpix2CartesianOperator(NSIDE)
    with pytest.warns(RuntimeWarning):
        op(pixel)


@pytest.mark.parametrize(
    'convention, dirs, dirs_za, shape',
    [
        ('azimuth,zenith', (0, 10), (10, 0), ()),
        ('azimuth,zenith', (0, 20), (20, 0), ()),
        ('azimuth,zenith', (0, 130), (130, 0), ()),
        ('azimuth,zenith', (20, 10), (10, 20), ()),
        ('azimuth,zenith', (130, 20), (20, 130), ()),
        (
            'azimuth,zenith',
            [(0, 10), (0, 20), (0, 130), (20, 10), (130, 20)],
            [(10, 0), (20, 0), (130, 0), (10, 20), (20, 130)],
            (5,),
        ),
        (
            'azimuth,zenith',
            [[(0, 10), (200, 20), (300, 130)]],
            [[(10, 0), (20, 200), (130, 300)]],
            (1, 3),
        ),
        ('elevation,azimuth', (80, 0), (10, 0), ()),
        ('elevation,azimuth', (70, 0), (20, 0), ()),
        ('elevation,azimuth', (-40, 0), (130, 0), ()),
        ('elevation,azimuth', (80, 20), (10, 20), ()),
        ('elevation,azimuth', (70, 130), (20, 130), ()),
        (
            'elevation,azimuth',
            [(80, 0), (70, 0), (-40, 0), (80, 20), (70, 130)],
            [(10, 0), (20, 0), (130, 0), (10, 20), (20, 130)],
            (5,),
        ),
        (
            'elevation,azimuth',
            [[(80, 0), (70, 200), (-40, 300)]],
            [[(10, 0), (20, 200), (130, 300)]],
            (1, 3),
        ),
        ('azimuth,elevation', (0, 80), (10, 0), ()),
        ('azimuth,elevation', (0, 70), (20, 0), ()),
        ('azimuth,elevation', (0, -40), (130, 0), ()),
        ('azimuth,elevation', (20, 80), (10, 20), ()),
        ('azimuth,elevation', (130, 70), (20, 130), ()),
        (
            'azimuth,elevation',
            [(0, 80), (0, 70), (0, -40), (20, 80), (130, 70)],
            [(10, 0), (20, 0), (130, 0), (10, 20), (20, 130)],
            (5,),
        ),
        (
            'azimuth,elevation',
            [[(0, 80), (200, 70), (300, -40)]],
            [[(10, 0), (20, 200), (130, 300)]],
            (1, 3),
        ),
    ],
)
def test_spherical_healpix(convention, dirs, dirs_za, shape):
    op_ref = Spherical2HealpixOperator(NSIDE, 'zenith,azimuth')
    refs = op_ref(np.radians(dirs_za))

    s2h = Spherical2HealpixOperator(NSIDE, convention)
    h2s = Healpix2SphericalOperator(NSIDE, convention)
    assert_allclose(s2h(np.radians(dirs)), refs)
    a = np.degrees(h2s(s2h(np.radians(dirs))))
    assert_equal(a.shape, shape + (2,))
    assert_allclose(a, dirs, rtol=1e-2, atol=0.5)


@pytest.mark.parametrize('convention1', Spherical2HealpixOperator.CONVENTIONS)
@pytest.mark.parametrize('convention2', Spherical2HealpixOperator.CONVENTIONS)
@pytest.mark.parametrize('nside', [NSIDE, 2 * NSIDE])
@pytest.mark.parametrize('nest', [False, True])
def test_spherical_healpix_rules(convention1, convention2, nside, nest):
    op = Spherical2HealpixOperator(NSIDE, convention1, nest=True)
    assert type(op.I) is Healpix2SphericalOperator
    assert op.I.convention == op.convention

    op2 = Healpix2SphericalOperator(nside, convention2, nest=nest)
    if convention2 == op.convention and nside == op.nside and nest == op.nest:
        assert type(op(op2)) is IdentityOperator
    else:
        assert type(op(op2)) is CompositionOperator


def test_spherical_healpix_invalid_convention():
    with pytest.raises(ValueError):
        Spherical2HealpixOperator(NSIDE, 'bla')


@pytest.mark.parametrize('shapein', [(), (1,), (2,), (3,)])
@pytest.mark.parametrize('shapeout', [(), (1,), (2,), (3,)])
def test_spherical_healpix_invalid_shape(shapein, shapeout):
    op = Spherical2HealpixOperator(NSIDE, 'zenith,azimuth')
    input = np.zeros(shapein)
    output = np.zeros(shapeout)

    if shapein == (2,) and shapeout == ():
        op(input, output)
        return

    with pytest.raises(ValueError):
        op(input, output)


@pytest.mark.parametrize('convention', Healpix2SphericalOperator.CONVENTIONS)
@pytest.mark.parametrize('nest', [False, True])
@pytest.mark.parametrize(
    'vec, shape',
    [
        (0, ()),
        (100, ()),
        (12 * NSIDE**2 - 1, ()),
        ([0, 100, 12 * NSIDE**2 - 1], (3,)),
        ([[0, 100]], (1, 2)),
    ],
)
def test_healpix_spherical(convention, nest, vec, shape):
    h2s = Healpix2SphericalOperator(NSIDE, convention, nest=nest)
    s2h = Spherical2HealpixOperator(NSIDE, convention, nest=nest)
    a = s2h(h2s(vec))
    assert a.shape == shape
    assert_allclose(a, vec)


@pytest.mark.parametrize('convention1', Healpix2SphericalOperator.CONVENTIONS)
@pytest.mark.parametrize('convention2', Healpix2SphericalOperator.CONVENTIONS)
@pytest.mark.parametrize('nside', [NSIDE, 2 * NSIDE])
@pytest.mark.parametrize('nest', [False, True])
def test_healpix_spherical_rules(convention1, convention2, nside, nest):
    op = Healpix2SphericalOperator(NSIDE, convention1, nest=True)
    assert type(op.I) is Spherical2HealpixOperator
    assert op.I.convention == op.convention

    op2 = Spherical2HealpixOperator(nside, convention2, nest=nest)
    if convention2 == op.convention and nside == op.nside and nest == op.nest:
        assert type(op(op2)) is IdentityOperator
    else:
        assert type(op(op2)) is CompositionOperator


def test_healpix_spherical_invalid_convention():
    with pytest.raises(ValueError):
        Healpix2SphericalOperator(NSIDE, 'bla')


@pytest.mark.parametrize('convention', Healpix2SphericalOperator.CONVENTIONS)
@pytest.mark.parametrize('shapein', [(), (1,), (2,), (3,)])
@pytest.mark.parametrize('shapeout', [(), (1,), (2,), (3,)])
def test_healpix_spherical_invalid_shapes(convention, shapein, shapeout):
    op = Healpix2SphericalOperator(NSIDE, 'zenith,azimuth')
    input = np.zeros(shapein)
    output = np.zeros(shapeout)

    if shapein == () and shapeout == (2,):
        op(input, output)
        return

    with pytest.raises(ValueError):
        op(input, output)


@pytest.mark.parametrize('convention', Healpix2SphericalOperator.CONVENTIONS)
@pytest.mark.parametrize('pixel', [-1, 12 * NSIDE**2])
def test_healpix_spherical_invalid_pixel(convention, pixel):
    op = Healpix2SphericalOperator(NSIDE, convention)
    with pytest.warns(RuntimeWarning):
        op(pixel)


def test_healpix_convolution():
    nside = 16
    keywords = {'fwhm': np.radians(30), 'iter': 2, 'lmax': 8, 'use_weights': False}

    input = np.arange(12 * nside**2)
    op = HealpixConvolutionGaussianOperator(**keywords)

    for i in input, np.repeat(input[:, None], 3, 1):
        expected = np.transpose(hp.smoothing(i.T, **keywords))
        assert_same(op(i), expected)

    if hp.__version__ <= '1.8.6':  # healpy #259
        return

    op = HealpixConvolutionGaussianOperator(pol=False, **keywords)
    input_ = np.arange(12 * nside**2)
    input = np.array([input_, input_, input_]).T
    expected_ = hp.smoothing(input_, **keywords)
    expected = np.array([expected_, expected_, expected_]).T
    assert_same(op(input), expected)


def test_healpix_convolution_morph():
    op = HealpixConvolutionGaussianOperator(fwhm=0)
    assert type(op) is IdentityOperator
    op = HealpixConvolutionGaussianOperator(sigma=0)
    assert type(op) is IdentityOperator


@pytest.mark.parametrize(
    'keywords', [{'iter': 2}, {'fwhm': -1}, {'sigma': -1}, {'fwhm': 1, 'sigma': 1}]
)
def test_healpix_convolution_errors(keywords):
    with pytest.raises(ValueError):
        HealpixConvolutionGaussianOperator(**keywords)


def test_healpix_laplacian():
    nside = 1
    ipix = 4
    npix = 12
    map = np.zeros(npix)
    map[ipix] = 1
    L = HealpixLaplacianOperator(nside)
    assert L.flags.square
    assert L.flags.symmetric
    h2 = np.array(4 * np.pi / npix)
    expected = [1, 0, 0, 1, -20, 4, 0, 4, 1, 0, 0, 1] / (6 * h2)
    assert_same(L(map), expected)
    assert_same(
        L(np.repeat(map, 5).reshape(12, 5)), expected[:, None], broadcasting=True
    )



import numpy as np

from ..geometry import create_circle, create_grid, create_grid_squares
from ..quantities import Quantity
from .core import PackedTable

__all__ = ['Layout', 'LayoutGrid', 'LayoutGridSquares']


class Layout(PackedTable):
    """
    A class for packed table including spatial information of the components.

    Attributes
    ----------
    shape : tuple of integers
        Tuple containing the layout dimensions.
    ndim : int
        The actual number of layout dimensions.
    center : array of shape (ncomponents, 2)
        Position of the component centers. The last dimension stands for
        the (X, Y) coordinates.
    vertex : array of shape (ncomponents, nvertices, 2), optional
        Position of the component vertices. This attribute is not present if
        it has not been specified. The last dimension stands for
        the (X, Y) coordinates.
    nvertices : int, optional
        Number of vertices if the vertex attribute is present.
    all : object
        Access to the unpacked attributes.

    """

    def __init__(self, shape, **keywords):
        """
        shape : tuple of int
            The layout unpacked shape. For 2-dimensional layouts, the shape would
            be (nrows, ncolumns).
        ndim : int, optional
            The number of splittable (indexable) dimensions. It is the actual
            number of dimensions of the layout. It can be lower than that
            specified by the layout shape, in which case the extra dimensions
            are instructed not to be split.
        selection : array-like of bool or int, slices, optional
            The slices or the integer or boolean selection that specifies
            the selected components (and reject those that are not physically
            present or those not handled by the current MPI process when the
            layout is distributed in a parallel processing).
        ordering : array-like of int, optional
            The values in this array specify an ordering of the components. It is
            used to define the 1-dimensional indexing of the packed components.
            A negative value means that the component is removed.
        center : array-like, Quantity, optional
            The position of the components. For 2-dimensional layouts, the expected
            shape is (nrows, ncolumns, 2). If not provided, these positions are
            derived from the vertex keyword, if present.
        vertex : array-like, Quantity, optional
            The position of the component vertices. For 2-dimensional layouts,
            the expected shape is (nrows, ncolumns, nvertices, 2). The last
            dimension stands for the (X, Y) coordinates.

        """
        if hasattr(self, 'vertex') and 'vertex' not in keywords:
            keywords['vertex'] = None
        if hasattr(self, 'center') and 'center' not in keywords:
            keywords['center'] = None
        if 'center' not in keywords and 'vertex' in keywords:
            keywords['center'] = lambda s: np.mean(s.vertex, axis=-2)
        if 'center' not in keywords:
            raise ValueError('The spatial layout is not defined.')
        PackedTable.__init__(self, shape, **keywords)
        if hasattr(self, 'vertex'):
            self.nvertices = self.vertex.shape[-2]

    def plot(
        self,
        autoscale=True,
        transform=None,
        edgecolor=None,
        facecolor=None,
        fill=None,
        **keywords,
    ):
        """
        Plot the layout.

        Parameters
        ----------
        autoscale : boolean
            If true, the axes of the plot will be updated to match the
            boundaries of the detectors.
        transform : callable, Operator
            Operator to be used to transform the layout coordinates into
            the data coordinate system.

        """
        import matplotlib.pyplot as mp

        if hasattr(self, 'vertex'):
            coords = self.vertex[..., :2]
        elif hasattr(self, 'radius'):
            coords = create_circle(self.radius, center=self.center[..., :2])
        else:
            coords = self.center[..., :2]

        if transform is not None:
            coords = transform(coords)

        if coords.ndim == 3:
            if fill is None and facecolor is not None:
                fill = True
            a = mp.gca()
            try:
                from pyoperators.utils import zip_broadcast

                def isnumber(x):
                    try:
                        float(x)
                        return True
                    except (TypeError, ValueError):
                        return False

                # special treatment for RGB tuples
                if isinstance(edgecolor, tuple) and all(isnumber(_) for _ in edgecolor):
                    edgecolor = [edgecolor]
                if isinstance(facecolor, tuple) and all(isnumber(_) for _ in facecolor):
                    facecolor = [facecolor]
                for c, ec, fc in zip_broadcast(
                    coords, edgecolor, facecolor, iter_str=False
                ):
                    a.add_patch(
                        mp.Polygon(
                            c,
                            closed=True,
                            edgecolor=ec,
                            facecolor=fc,
                            fill=fill,
                            **keywords,
                        )
                    )
            except ImportError:
                # for PyOperators < 0.13.1
                for c in coords:
                    a.add_patch(
                        mp.Polygon(
                            c,
                            closed=True,
                            edgecolor=edgecolor,
                            facecolor=facecolor,
                            fill=fill,
                            **keywords,
                        )
                    )
            if autoscale:
                a.autoscale_view()
        elif coords.ndim == 2:
            if 'color' not in keywords:
                keywords['color'] = 'black'
            if 'marker' not in keywords:
                keywords['marker'] = 'o'
            if 'linestyle' not in keywords:
                keywords['linestyle'] = ''
            mp.plot(
                coords[:, 0],
                coords[:, 1],
                scalex=autoscale,
                scaley=autoscale,
                **keywords,
            )
        else:
            raise ValueError('Invalid number of dimensions.')
        mp.show()


class LayoutGrid(Layout):
    """
    The components are laid out in a rectangular grid (nrows, ncolumns).
    Their positions are expressed in the (X, Y) referential.

       Y ^
         |
         +--> X

    Before rotation, rows increase along the -Y axis (unless yreflection is
    set to True) and columns increase along the X axis (unless xreflection
    is set to True).

    Example
    -------
    >>> spacing = 0.001
    >>> layout = LayoutGrid((16, 16), spacing)
    >>> layout.plot()

    Attributes
    ----------
    shape : tuple of integers (nrows, ncolumns)
        Tuple containing the number of rows and columns of the grid.
    ndim : int
        The number of layout dimensions.
    center : array of shape (len(self), 2)
        Position of the component centers. The last dimension stands for
        the (X, Y) coordinates.
    all : object
        Access to the unpacked attributes of shape (nrows, ncolumns, ...)

    """

    def __init__(
        self,
        shape,
        spacing,
        xreflection=False,
        yreflection=False,
        angle=0,
        origin=(0, 0),
        startswith1=False,
        _z=None,
        **keywords,
    ):
        """
        shape : tuple of two integers (nrows, ncolumns)
            Number of rows and columns of the grid.
        spacing : float, Quantity
            Physical spacing between components.
        xreflection : boolean, optional
            Reflection along the X-axis (before rotation).
        yreflection : boolean, optional
            Reflection along the Y-axis (before rotation).
        angle : float, optional
            Counter-clockwise rotation angle in degrees (before translation).
        origin : array-like of shape (2,), optional
            The (X, Y) coordinates of the grid center
        startswith1 : boolean, optional
            If True, start column and row indewing with one.
        _z : float, optional, don't use it yet
            The third dimension of the component' centers or vertices.
        selection : array-like of bool or int, slices, optional
            The slices or the integer or boolean selection that specifies
            the selected components (and reject those that are not physically
            present or those not handled by the current MPI process when the
            layout is distributed in a parallel processing).
        ordering : array-like of int, optional
            The values in this array specify an ordering of the components. It is
            used to define the 1-dimensional indexing of the packed components.
            A negative value means that the component is removed.

        """
        if not isinstance(shape, (list, tuple)):
            raise TypeError('Invalid grid shape.')
        if len(shape) != 2:
            raise ValueError('The layout grid is not 2-dimensional.')
        if len(origin) != len(shape):
            raise ValueError('Invalid dimensions of the layout center.')

        unit = (
            getattr(spacing, 'unit', '')
            or getattr(origin, 'unit', '')
            or 'radius' in keywords
            and getattr(keywords['radius'], 'unit', '')
        )
        if unit:
            spacing = Quantity(spacing).tounit(unit)
            origin = Quantity(origin).tounit(unit)
            if 'radius' in keywords:
                keywords['radius'] = Quantity(keywords['radius']).tounit(unit)
        if 'vertex' not in keywords:
            center = create_grid(
                shape,
                spacing,
                xreflection=xreflection,
                yreflection=yreflection,
                center=origin,
                angle=angle,
            )
            if _z is not None:
                center = np.concatenate([center, np.full_like(center[..., :1], _z)], -1)
            keywords['center'] = center
        if unit:
            if 'center' in keywords:
                keywords['center'] = Quantity(keywords['center'], unit, copy=False)
            if 'vertex' in keywords:
                keywords['vertex'] = Quantity(keywords['vertex'], unit, copy=False)
        self.spacing = spacing
        self.origin = origin
        self.xreflection = xreflection
        self.yreflection = yreflection
        self.angle = angle
        self.startswith1 = startswith1
        self.unit = unit
        Layout.__init__(self, shape, **keywords)
        self._special_attributes += ('column', 'row')

    def column(self):
        return self.index % self.shape[1] + self.startswith1

    def row(self):
        return self.index // self.shape[1] + self.startswith1


class LayoutGridSquares(LayoutGrid):
    """
    The square components are laid out in a rectangular grid (nrows, ncolumns).
    The positions of their corners are expressed in the (X, Y) referential.

       Y ^
         |
         +--> X

    Before rotation, rows increase along the -Y axis (unless yreflection is
    set to True) and columns increase along the X axis (unless xreflection
    is set to True).

    Example
    -------
    >>> spacing = 0.001
    >>> layout = LayoutGridSquares((8, 8), spacing, filling_factor=0.8)
    >>> layout.plot()

    Attributes
    ----------
    shape : tuple of integers (nrows, ncolumns)
        See below.
    ndim : int
        The number of layout dimensions.
    center : array of shape (len(self), 2)
        Position of the component centers. The last dimension stands for
        the (X, Y) coordinates.
    vertex : array of shape (len(self), 4, 2)
        Corners of the components. The dimensions refers to the component
        number, the corner number counted counter-clockwise starting from
        the top-right and the (X, Y) coordinates.
    all : object
        Access to the unpacked attributes of shape (nrows, ncolumns, ...)

    """

    def __init__(
        self,
        shape,
        spacing,
        filling_factor=1,
        xreflection=False,
        yreflection=False,
        angle=0,
        origin=(0, 0),
        startswith1=False,
        **keywords,
    ):
        """
        shape : tuple of two integers (nrows, ncolumns)
            Number of rows and columns of the grid.
        spacing : float, Quantity
            Physical spacing between components.
        filling_factor : float
            Ratio of the component area over spacing**2.
        xreflection : boolean
            Reflection along the X-axis (before rotation).
        yreflection : boolean
            Reflection along the Y-axis (before rotation).
        angle : float
            Counter-clockwise rotation angle in degrees (before translation).
        origin : array-like of shape (2,)
            The (X, Y) coordinates of the grid center
        startswith1 : boolean, optional
            If True, start column and row indewing with one.
        selection : array-like of bool or int, slices, optional
            The slices or the integer or boolean selection that specifies
            the selected components (and reject those that are not physically
            present or those not handled by the current MPI process when the
            layout is distributed in a parallel processing).
        ordering : array-like of int, optional
            The values in this array specify an ordering of the components. It is
            used to define the 1-dimensional indexing of the packed components.
            A negative value means that the component is removed.

        """
        unit = getattr(spacing, 'unit', '') or getattr(origin, 'unit', '')
        if unit:
            spacing = Quantity(spacing).tounit(unit)
            origin = Quantity(origin).tounit(unit)
        vertex = create_grid_squares(
            shape,
            spacing,
            filling_factor=filling_factor,
            xreflection=xreflection,
            yreflection=yreflection,
            center=origin,
            angle=angle,
        )
        if unit:
            vertex = Quantity(vertex, unit, copy=False)
        LayoutGrid.__init__(
            self,
            shape,
            spacing,
            vertex=vertex,
            xreflection=xreflection,
            yreflection=yreflection,
            angle=angle,
            origin=origin,
            startswith1=startswith1,
            **keywords,
        )
        self.filling_factor = filling_factor



import itertools

import numpy as np
import pytest
import scipy.sparse

import pyoperators
from pyoperators import (
    CompositionOperator,
    DiagonalOperator,
    Rotation2dOperator,
    Rotation3dOperator,
)
from pyoperators.utils.testing import assert_same
from pysimulators.sparse import (
    FSCMatrix,
    FSCRotation2dMatrix,
    FSCRotation3dMatrix,
    FSRBlockMatrix,
    FSRMatrix,
    FSRRotation2dMatrix,
    FSRRotation3dMatrix,
    SparseOperator,
)

from .common import BIGGEST_FLOAT_TYPE, FLOAT_TYPES, INT_TYPES, SINT_TYPES


def min_type(t1, t2):
    return min(np.dtype(t1), np.dtype(t2)).type


@pytest.mark.parametrize('itype', INT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize('vtype', FLOAT_TYPES)
@pytest.mark.parametrize('block_size', [1, 2])
def test_fsc1(itype, ftype, vtype, block_size):
    input = [1, 2, 1, 1, 1, 1]
    index = [3, 2, 2, 1, 2, -1]
    index_u = [3, 2, 2, 1, 2, 3]
    value = [1, 1, 0.5, 1, 2, 10]
    expected = [0, 1, 4.5, 1]
    expected_u = [0, 1, 4.5, 11]

    if np.dtype(itype).kind != 'u':
        ind = index
        exp = expected
    else:
        ind = index_u
        exp = expected_u
    input_ = np.array(input, vtype)
    if block_size == 2:
        input_ = np.array([input_, input_]).T.ravel()
        exp = np.array([exp, exp]).T.ravel()
    dtype = [('index', itype), ('value', ftype)]
    matrix = np.recarray((6, 1), dtype=dtype)
    matrix[..., 0].index = ind
    matrix[..., 0].value = value
    op = FSCMatrix((4, 6), data=matrix)
    out = op * np.array(input_, vtype)
    assert_same(out, exp)
    out[...] = 0
    op._matvec(np.array(input_, vtype), out=out)
    assert_same(out, exp)


def get_fsc2_mat(itype, ftype):
    if np.dtype(itype).kind != 'u':
        index = [[3, 2, 2, 1, 2, -1], [-1, 3, 1, 1, 0, -1]]
    else:
        index = [[3, 2, 2, 1, 2, 3], [3, 3, 1, 1, 0, 3]]
    value = [[1, 1, 0.5, 1, 2, 10], [1, 2, 0.5, 1, 1, 10]]
    dtype = [('index', itype), ('value', ftype)]
    matrix = np.recarray((6, 2), dtype=dtype)
    matrix[..., 0].index, matrix[..., 1].index = index
    matrix[..., 0].value, matrix[..., 1].value = value
    return FSCMatrix((4, 6), data=matrix)


@pytest.mark.parametrize('itype', INT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize('vtype', FLOAT_TYPES)
@pytest.mark.parametrize('block_size', [1, 2])
def test_fsc2(itype, ftype, vtype, block_size):
    input_fsc = np.asarray([1, 2, 1, 1, 1, 1], vtype)
    input_fsr = np.asarray([1, 2, 1, 4], vtype)
    if np.dtype(itype).kind != 'u':
        expected = np.asarray([1, 2.5, 4.5, 5])
    else:
        expected = np.asarray([1, 2.5, 4.5, 26])

    if block_size == 2:
        input_fsc = np.array([input_fsc, input_fsc]).T.ravel()
        input_fsr = np.array([input_fsr, input_fsr]).T.ravel()
        expected = np.array([expected, expected]).T.ravel()

    mat = get_fsc2_mat(itype, ftype)
    out = mat * input_fsc
    assert_same(out, expected)
    out[...] = 0
    mat._matvec(input_fsc, out=out)
    assert_same(out, expected)
    out = input_fsr * mat
    assert_same(out, FSRMatrix(mat.shape[::-1], data=mat.data) * input_fsr)
    out = (3 * mat) * input_fsc
    assert_same(out, 3 * expected)
    out = (mat * 3) * input_fsc
    assert_same(out, 3 * expected)


@pytest.mark.parametrize('itype', INT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize('shapein', [(), (3,)])
def test_fsc2_dense(itype, ftype, shapein):
    mat = get_fsc2_mat(itype, ftype)
    op = SparseOperator(mat)
    todense = op.todense(shapein=(6,) + shapein)
    assert_same(todense.T, op.T.todense(shapeout=(6,) + shapein))
    op2 = SparseOperator(mat, shapeout=(2, 2) + shapein, shapein=(3, 2) + shapein)
    assert_same(op2.todense(), todense)


def test_fsc3():
    mat = FSCMatrix((3, 4), 10)
    assert mat.data.shape == (4, 10)
    assert mat.data.index.dtype == int
    assert mat.data.value.dtype == float

    mat = FSCMatrix((3, 4), 10, dtype=np.float16, dtype_index=np.int8)
    assert mat.data.index.dtype == np.int8
    assert mat.data.value.dtype == np.float16


def test_fsc_error():
    data = np.zeros((3, 4), [('index', int), ('value', float)])
    data_wrong1 = np.empty((3, 4), [('index_', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('value', float), ('index', int)])

    with pytest.raises(TypeError):
        FSCMatrix(3)

    with pytest.raises(ValueError):
        FSCMatrix((2, 3, 4))

    with pytest.raises(ValueError):
        FSCMatrix((3, 8), data=data)

    with pytest.raises(TypeError):
        FSCMatrix((8, 3), data=data_wrong1)

    with pytest.raises(TypeError):
        FSCMatrix((8, 3), data=data_wrong2)

    with pytest.raises(ValueError):
        FSCMatrix((8, 3))

    mat = FSCMatrix((8, 3), data=data)
    mat._matvec(np.ones(3))
    with pytest.raises(ValueError):
        mat._matvec(np.ones(7))

    with pytest.raises(TypeError):
        mat._matvec(np.ones(3), out=1)

    with pytest.raises(ValueError):
        mat._matvec(np.ones(3), out=np.zeros(7))


def get_fsr1_mat(itype, ftype):
    if np.dtype(itype).kind != 'u':
        ind = [3, 2, 2, 1, 2, -1]
    else:
        ind = [3, 2, 2, 1, 2, 3]
    value = [1, 1, 0.5, 1, 2, 10]
    dtype = [('index', itype), ('value', ftype)]

    matrix = np.recarray((6, 1), dtype=dtype)
    matrix[..., 0].index = ind
    matrix[..., 0].value = value
    return FSRMatrix((6, 4), data=matrix)


@pytest.mark.parametrize('itype', INT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize('vtype', FLOAT_TYPES)
@pytest.mark.parametrize('block_size', [1, 2])
def test_fsr1(itype, ftype, vtype, block_size):
    input = [1, 2, 3, 4]
    expected = [4, 3, 1.5, 2, 6, 0]
    expected_u = [4, 3, 1.5, 2, 6, 40]

    if np.dtype(itype).kind != 'u':
        exp = expected
    else:
        exp = expected_u
    input_ = np.array(input, vtype)
    if block_size == 2:
        input_ = np.array([input_, input_]).T.ravel()
        exp = np.array([exp, exp]).T.ravel()
    op = get_fsr1_mat(itype, ftype)
    out = op * input_
    assert_same(out, exp)
    out[...] = 0
    op._matvec(input_, out=out)
    assert_same(out, exp)


@pytest.mark.parametrize('itype', INT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
def test_fsr1_dense(itype, ftype):
    mat = get_fsr1_mat(itype, ftype)
    op = SparseOperator(mat, shapein=4)
    todense = op.todense()
    pTp = op.T * op
    if (itype, ftype) not in (
        (np.int32, np.float32),
        (np.int32, np.float64),
        (np.int64, np.float32),
        (np.int64, np.float64),
    ):
        assert type(pTp) is CompositionOperator
        return
    assert type(pTp) is DiagonalOperator
    assert_same(pTp.todense(), np.dot(todense.T, todense))


def get_fsr2_mat(itype, ftype):
    if np.dtype(itype).kind != 'u':
        index = [[3, 2, 2, 1, 2, -1], [-1, 3, 1, 1, 0, -1]]
    else:
        index = [[3, 2, 2, 1, 2, 3], [3, 3, 1, 1, 0, 3]]
    value = [[1, 1, 0.5, 1, 2, 10], [1, 2, 0.5, 1, 1, 10]]
    dtype = [('index', itype), ('value', ftype)]
    matrix = np.recarray((6, 2), dtype=dtype)
    matrix[..., 0].index, matrix[..., 1].index = index
    matrix[..., 0].value, matrix[..., 1].value = value
    return FSRMatrix((6, 4), data=matrix)


@pytest.mark.parametrize('itype', INT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize('vtype', FLOAT_TYPES)
@pytest.mark.parametrize('block_size', [1, 2])
def test_fsr2(itype, ftype, vtype, block_size):
    input_fsr = [1, 2, 3, 4]
    input_fsc = [1, 2, 3, 4, 5, 6]
    expected = [4, 11, 2.5, 4, 7, 0]
    expected_u = [8, 11, 2.5, 4, 7, 80]

    input_fsc_ = np.asarray(input_fsc, vtype)
    input_fsr_ = np.asarray(input_fsr, vtype)
    if np.dtype(itype).kind != 'u':
        exp = np.asarray(expected)
    else:
        exp = np.asarray(expected_u)
    if block_size == 2:
        input_fsc_ = np.array([input_fsc_, input_fsc_]).T.ravel()
        input_fsr_ = np.array([input_fsr_, input_fsr_]).T.ravel()
        exp = np.array([exp, exp]).T.ravel()
    mat = get_fsr2_mat(itype, ftype)
    out = mat * input_fsr_
    assert_same(out, exp)
    out[...] = 0
    mat._matvec(input_fsr_, out=out)
    assert_same(out, exp)
    out = input_fsc_ * mat
    assert_same(out, FSCMatrix(mat.shape[::-1], data=mat.data) * input_fsc_)
    out = (3 * mat) * input_fsr_
    assert_same(out, 3 * exp)
    out = (mat * 3) * input_fsr_
    assert_same(out, 3 * exp)


@pytest.mark.parametrize('itype', INT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize('shapein', [(), (3,)])
def test_fsr2_dense(itype, ftype, shapein):
    mat = get_fsr2_mat(itype, ftype)
    op = SparseOperator(mat)
    todense = op.todense(shapein=(4,) + shapein)
    assert_same(todense.T, op.T.todense(shapeout=(4,) + shapein))
    if shapein is None:
        return
    op2 = SparseOperator(mat, shapein=(2, 2) + shapein, shapeout=(3, 2) + shapein)
    assert_same(op2.todense(), todense)


def test_fsr3():
    mat = FSRMatrix((4, 3), 10)
    assert mat.data.shape == (4, 10)
    assert mat.data.index.dtype == int
    assert mat.data.value.dtype == float

    mat = FSRMatrix((4, 3), 10, dtype=np.float16, dtype_index=np.int8)
    assert mat.data.index.dtype == np.int8
    assert mat.data.value.dtype == np.float16


def test_fsr_error():
    data = np.zeros((3, 4), [('index', int), ('value', float)])
    data_wrong1 = np.empty((3, 4), [('index_', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('value', float), ('index', int)])

    with pytest.raises(TypeError):
        FSRMatrix(3)

    with pytest.raises(ValueError):
        FSRMatrix((2, 3, 4))

    with pytest.raises(ValueError):
        FSRMatrix((8, 3), data=data)

    with pytest.raises(TypeError):
        FSRMatrix((3, 8), data=data_wrong1)

    with pytest.raises(TypeError):
        FSRMatrix((3, 8), data=data_wrong2)

    with pytest.raises(ValueError):
        FSRMatrix((3, 8), 5, data=data)

    with pytest.raises(ValueError):
        FSRMatrix((3, 8))

    mat = FSRMatrix((3, 8), data=data)
    mat._matvec(np.ones(8))
    with pytest.raises(ValueError):
        mat._matvec(np.ones(7))

    with pytest.raises(TypeError):
        mat._matvec(np.ones(8), out=1)

    with pytest.raises(ValueError):
        mat._matvec(np.ones(8), out=np.zeros(4))


def get_block_mat(itype, ftype, inner_block_shape):
    outer_block_shape = (6, 4)
    indices = [[3, 2, 2, 1, 2, -1], [-1, 3, 1, 1, 0, -1]]
    shape = [i * o for i, o in zip(inner_block_shape, outer_block_shape)]
    mat = FSRBlockMatrix(
        shape, inner_block_shape, len(indices), dtype=ftype, dtype_index=itype
    )
    for k, index in enumerate(indices):
        mat.data[:, k].index = index
    for i in range(outer_block_shape[0]):
        for k in range(len(indices)):
            mat.data[i, k]['value'][...] = np.random.randint(-10, 11, inner_block_shape)

    dense = np.zeros(shape, dtype=ftype)
    for i, d in enumerate(mat.data):
        a1, z1 = i * inner_block_shape[0], (i + 1) * inner_block_shape[0]
        for d_ in d:
            j = d_['index']
            if j < 0:
                continue
            a2, z2 = j * inner_block_shape[1], (j + 1) * inner_block_shape[1]
            dense[a1:z1, a2:z2] += d_['value']
    return mat, dense


@pytest.mark.parametrize('itype', SINT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
@pytest.mark.parametrize(
    'inner_block_shape',
    itertools.islice(itertools.product([1, 2, 3], repeat=2), 1, None),
)
def test_block(itype, ftype, inner_block_shape):
    np.random.seed(0)

    mat_fsr, dense = get_block_mat(itype, ftype, inner_block_shape)
    mat_fsc = mat_fsr._transpose()
    input_fsc = np.random.random_sample(mat_fsc.shape[1])
    input_fsr = np.random.random_sample(mat_fsr.shape[1])

    assert_same(mat_fsr.block_shape, inner_block_shape)
    assert_same(mat_fsc.block_shape, inner_block_shape[::-1])
    op = SparseOperator(mat_fsc)
    assert op.matrix.dtype == ftype
    assert op.dtype == ftype
    assert_same(op.todense(), dense.T)
    assert_same(op.T.todense(), dense)

    ref = mat_fsc._matvec(input_fsc.astype(BIGGEST_FLOAT_TYPE))
    for ftype2 in FLOAT_TYPES:
        out = np.zeros_like(ref, ftype2)
        mat_fsc._matvec(input_fsc.astype(ftype2), out)
        assert_same(out, ref.astype(min_type(ftype, ftype2)), atol=100)

    ref = (2 * (mat_fsc * input_fsc)).astype(ftype)
    assert_same((mat_fsc * 2) * input_fsc, ref, atol=100)
    assert_same((2 * mat_fsc) * input_fsc, ref, atol=100)
    assert_same(input_fsr * mat_fsc, mat_fsr * input_fsr, atol=10)

    op = SparseOperator(mat_fsr)
    assert op.matrix.dtype == ftype
    assert op.dtype == ftype
    assert_same(op.todense(), dense)
    assert_same(op.T.todense(), dense.T)

    ref = mat_fsr._matvec(input_fsr.astype(BIGGEST_FLOAT_TYPE))
    for ftype2 in FLOAT_TYPES:
        out = np.zeros_like(ref, ftype2)
        mat_fsr._matvec(input_fsr.astype(ftype2), out)
        assert_same(out, ref.astype(min_type(ftype, ftype2)), atol=10)

    ref = (2 * (mat_fsr * input_fsr)).astype(ftype)
    assert_same((mat_fsr * 2) * input_fsr, ref, atol=10)
    assert_same((2 * mat_fsr) * input_fsr, ref, atol=10)
    assert_same(input_fsc * mat_fsr, mat_fsc * input_fsc, atol=10)


def fill_rot2d(array, dense, index, value, angle, n):
    for i, (j, v, a) in enumerate(zip(index, value, angle)):
        array[i, n].index = j
        if j == -1:
            array[i, n].r11 = 0
            array[i, n].r21 = 0
            continue
        r = v * Rotation2dOperator(a, degrees=True).todense(shapein=2)
        array[i, n].r11 = r[0, 0]
        array[i, n].r21 = r[1, 0]
        dense[2 * i : 2 * i + 2, 2 * j : 2 * j + 2] += r


@pytest.mark.parametrize('itype', SINT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
def test_rot2d(itype, ftype):
    input_fsc = np.arange(6 * 2, dtype=ftype)
    input_fsr = np.arange(4 * 2, dtype=ftype)
    index1 = [3, 2, 2, 1, 2, -1]
    index2 = [-1, 3, 1, 1, 0, -1]
    value1 = [1, 1, 0.5, 1, 2, 10]
    value2 = [1, 2, 0.5, 1, 1, 10]
    angle1 = [1, 10, 20, 30, 40, -10]
    angle2 = [0, 30, -20, 1, 1, 10]

    array = np.recarray(
        (6, 2), dtype=[('index', itype), ('r11', ftype), ('r21', ftype)]
    )
    dense = np.zeros((6 * 2, 4 * 2), dtype=ftype)
    fill_rot2d(array, dense, index1, value1, angle1, 0)
    fill_rot2d(array, dense, index2, value2, angle2, 1)

    mat_fsc = FSCRotation2dMatrix(dense.shape[::-1], data=array)
    mat_fsr = FSRRotation2dMatrix(dense.shape, data=array)

    op = SparseOperator(mat_fsc)
    assert op.matrix.dtype == ftype
    assert op.dtype == ftype
    assert_same(op.todense(), dense.T)
    assert_same(op.T.todense(), dense)
    ref = mat_fsc._matvec(input_fsc.astype(BIGGEST_FLOAT_TYPE))
    for ftype2 in FLOAT_TYPES:
        out = np.zeros_like(ref, ftype2)
        mat_fsc._matvec(input_fsc.astype(ftype2), out)
        assert_same(out, ref.astype(min_type(ftype, ftype2)))
    ref = 2 * (mat_fsc * input_fsc)
    assert_same((mat_fsc * 2) * input_fsc, ref)
    assert_same((2 * mat_fsc) * input_fsc, ref)
    assert_same(input_fsr * mat_fsc, mat_fsr * input_fsr)

    op = SparseOperator(mat_fsr)
    assert op.matrix.dtype == ftype
    assert op.dtype == ftype
    assert_same(op.todense(), dense)
    assert_same(op.T.todense(), dense.T)
    ref = mat_fsr._matvec(input_fsr.astype(BIGGEST_FLOAT_TYPE))
    for ftype2 in FLOAT_TYPES:
        out = np.zeros_like(ref, ftype2)
        mat_fsr._matvec(input_fsr.astype(ftype2), out)
        assert_same(out, ref.astype(min_type(ftype, ftype2)))
    ref = 2 * (mat_fsr * input_fsr)
    assert_same((mat_fsr * 2) * input_fsr, ref)
    assert_same((2 * mat_fsr) * input_fsr, ref)
    assert_same(input_fsc * mat_fsr, mat_fsc * input_fsc)


@pytest.mark.parametrize('itype', SINT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
def test_rot2d_dense(itype, ftype):
    index1 = [3, 2, 2, 1, 2, -1]
    value1 = [1, 1, 0.5, 1, 2, 10]
    angle1 = [1, 10, 20, 30, 40, -10]

    array = np.recarray(
        (6, 1), dtype=[('index', itype), ('r11', ftype), ('r21', ftype)]
    )
    dense = np.zeros((6 * 2, 4 * 2), dtype=ftype)
    fill_rot2d(array, dense, index1, value1, angle1, 0)
    op = SparseOperator(FSRRotation2dMatrix(dense.shape, data=array))
    pTp = op.T * op
    if (itype, ftype) not in (
        (np.int32, np.float32),
        (np.int32, np.float64),
        (np.int64, np.float32),
        (np.int64, np.float64),
    ):
        assert type(pTp) is CompositionOperator
        return
    assert type(pTp) is DiagonalOperator
    assert_same(pTp.todense(), np.dot(dense.T, dense), atol=1)


def test_fsc_rot2d_error():
    data = np.zeros((3, 4), dtype=[('index', int), ('r11', float), ('r21', float)])
    data_wrong1 = np.empty((3, 4), [('index', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('r11', float), ('r21', float), ('index', int)])
    data_wrong3 = np.empty(
        (3, 4), dtype=[('index', int), ('r11', float), ('r21', np.float32)]
    )

    with pytest.raises(TypeError):
        FSCRotation2dMatrix(2)

    with pytest.raises(ValueError):
        FSCRotation2dMatrix((2, 3, 4))

    with pytest.raises(ValueError):
        FSCRotation2dMatrix((6, 16), data=data)

    with pytest.raises(ValueError):
        FSCRotation2dMatrix((7, 16), data=data)

    with pytest.raises(ValueError):
        FSCRotation2dMatrix((17, 6), data=data)

    with pytest.raises(TypeError):
        FSCRotation2dMatrix((16, 6), data=data_wrong1)

    with pytest.raises(TypeError):
        FSCRotation2dMatrix((16, 6), data=data_wrong2)

    with pytest.raises(TypeError):
        FSCRotation2dMatrix((16, 6), data=data_wrong3)

    with pytest.raises(ValueError):
        FSCRotation2dMatrix((16, 6), 5, data=data)

    with pytest.raises(ValueError):
        FSCRotation2dMatrix((16, 6))

    mat = FSCRotation2dMatrix((16, 6), data=data)
    with pytest.raises(ValueError):
        mat._matvec(np.ones(4))

    mat._matvec(np.ones((3, 2)))
    with pytest.raises(ValueError):
        mat._matvec(np.ones(17))

    with pytest.raises(ValueError):
        mat._matvec(np.ones((8, 2)))

    with pytest.raises(TypeError):
        mat._matvec(np.ones((3, 2)), out=1)

    with pytest.raises(ValueError):
        mat._matvec(np.ones((3, 2)), out=np.zeros(7))


def test_fsr_rot2d_error():
    data = np.zeros((3, 4), dtype=[('index', int), ('r11', float), ('r21', float)])
    data_wrong1 = np.empty((3, 4), [('index', int), ('value', float)])
    data_wrong2 = np.empty((3, 4), [('r11', float), ('r21', float), ('index', int)])
    data_wrong3 = np.empty(
        (3, 4), dtype=[('index', int), ('r11', float), ('r21', np.float32)]
    )

    with pytest.raises(TypeError):
        FSRRotation2dMatrix(3)

    with pytest.raises(ValueError):
        FSRRotation2dMatrix((2, 3, 4))

    with pytest.raises(ValueError):
        FSRRotation2dMatrix((16, 6), data=data)

    with pytest.raises(ValueError):
        FSRRotation2dMatrix((7, 16), data=data)

    with pytest.raises(ValueError):
        FSRRotation2dMatrix((6, 17), data=data)

    with pytest.raises(TypeError):
        FSRRotation2dMatrix((6, 16), data=data_wrong1)

    with pytest.raises(TypeError):
        FSRRotation2dMatrix((6, 16), data=data_wrong2)

    with pytest.raises(TypeError):
        FSRRotation2dMatrix((6, 16), data=data_wrong3)

    with pytest.raises(ValueError):
        FSRRotation2dMatrix((6, 16), 5, data=data)

    with pytest.raises(ValueError):
        FSRRotation2dMatrix((6, 16))

    mat = FSRRotation2dMatrix((6, 16), data=data)
    with pytest.raises(ValueError):
        mat._matvec(np.ones(4))

    mat._matvec(np.ones((8, 2)))
    with pytest.raises(ValueError):
        mat._matvec(np.ones(17))

    with pytest.raises(ValueError):
        mat._matvec(np.ones((3, 2)))

    with pytest.raises(TypeError):
        mat._matvec(np.ones((8, 2)), out=1)

    with pytest.raises(ValueError):
        mat._matvec(np.ones((8, 2)), out=np.zeros(7))


def fill_rot3d(array, dense, index, value, angle, n):
    for i, (j, v, a) in enumerate(zip(index, value, angle)):
        array[i, n].index = j
        if j == -1:
            array[i, n].r11 = 0
            array[i, n].r22 = 0
            array[i, n].r32 = 0
            continue
        r = v * Rotation3dOperator('X', a, degrees=True).todense(shapein=3)
        array[i, n].r11 = r[0, 0]
        array[i, n].r22 = r[1, 1]
        array[i, n].r32 = r[2, 1]
        dense[3 * i : 3 * i + 3, 3 * j : 3 * j + 3] += r


@pytest.mark.parametrize('itype', SINT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
def test_rot3d(itype, ftype):
    input_fsc = np.arange(6 * 3, dtype=ftype)
    input_fsr = np.arange(4 * 3, dtype=ftype)
    index1 = [3, 2, 2, 1, 2, -1]
    index2 = [-1, 3, 1, 1, 0, -1]
    value1 = [1, 1, 0.5, 1, 2, 10]
    value2 = [1, 2, 0.5, 1, 1, 10]
    angle1 = [1, 10, 20, 30, 40, -10]
    angle2 = [0, 30, -20, 1, 1, 10]

    array = np.recarray(
        (6, 2),
        dtype=[('index', itype), ('r11', ftype), ('r22', ftype), ('r32', ftype)],
    )
    dense = np.zeros((6 * 3, 4 * 3), dtype=ftype)
    fill_rot3d(array, dense, index1, value1, angle1, 0)
    fill_rot3d(array, dense, index2, value2, angle2, 1)

    mat_fsc = FSCRotation3dMatrix(dense.shape[::-1], data=array)
    mat_fsr = FSRRotation3dMatrix(dense.shape, data=array)

    op = SparseOperator(mat_fsc)
    assert op.matrix.dtype == ftype
    assert op.dtype == ftype
    assert_same(op.todense(), dense.T)
    assert_same(op.T.todense(), dense)
    ref = mat_fsc._matvec(input_fsc.astype(BIGGEST_FLOAT_TYPE))
    for ftype2 in FLOAT_TYPES:
        out = np.zeros_like(ref, ftype2)
        mat_fsc._matvec(input_fsc.astype(ftype2), out)
        assert_same(out, ref.astype(min_type(ftype, ftype2)))
    ref = 3 * (mat_fsc * input_fsc)
    assert_same((mat_fsc * 3) * input_fsc, ref)
    assert_same((3 * mat_fsc) * input_fsc, ref)
    assert_same(input_fsr * mat_fsc, mat_fsr * input_fsr)

    op = SparseOperator(mat_fsr)
    assert op.matrix.dtype == ftype
    assert op.dtype == ftype
    assert_same(op.todense(), dense)
    assert_same(op.T.todense(), dense.T)
    ref = mat_fsr._matvec(input_fsr.astype(BIGGEST_FLOAT_TYPE))
    for ftype2 in FLOAT_TYPES:
        out = np.zeros_like(ref, ftype2)
        mat_fsr._matvec(input_fsr.astype(ftype2), out)
        assert_same(out, ref.astype(min_type(ftype, ftype2)))
    ref = 3 * (mat_fsr * input_fsr)
    assert_same((mat_fsr * 3) * input_fsr, ref, rtol=10)
    assert_same((3 * mat_fsr) * input_fsr, ref, rtol=10)
    assert_same(input_fsc * mat_fsr, mat_fsc * input_fsc, rtol=10)


@pytest.mark.parametrize('itype', SINT_TYPES)
@pytest.mark.parametrize('ftype', FLOAT_TYPES)
def test_rot3d_dense(itype, ftype):
    index1 = [3, 2, 2, 1, 2, -1]
    value1 = [1, 1, 0.5, 1, 2, 10]
    angle1 = [1, 10, 20, 30, 40, -10]

    array = np.recarray(
        (6, 1),
        dtype=[('index', itype), ('r11', ftype), ('r22', ftype), ('r32', ftype)],
    )
    dense = np.zeros((6 * 3, 4 * 3), dtype=ftype)
    fill_rot3d(array, dense, index1, value1, angle1, 0)
    op = SparseOperator(FSRRotation3dMatrix(dense.shape, data=array))
    pTp = op.T * op
    if (itype, ftype) not in (
        (np.int32, np.float32),
        (np.int32, np.float64),
        (np.int64, np.float32),
        (np.int64, np.float64),
    ):
        assert type(pTp) is CompositionOperator
        return
    assert type(pTp) is DiagonalOperator
    assert_same(pTp.todense(), np.dot(dense.T, dense), atol=1)


def test_fsc_rot3d_error():
    data = np.zeros(
        (3, 4), dtype=[('index', int), ('r11', float), ('r22', float), ('r32', float)]
    )
    data_wrong1 = np.empty((3, 4), [('index', int), ('value', float)])
    data_wrong2 = np.empty(
        (3, 4), [('r11', float), ('r22', float), ('r32', float), ('index', int)]
    )
    data_wrong3 = np.empty(
        (3, 4),
        dtype=[('index', int), ('r11', float), ('r22', np.float32), ('r32', float)],
    )
    with pytest.raises(TypeError):
        FSCRotation3dMatrix(3)

    with pytest.raises(ValueError):
        FSCRotation3dMatrix((2, 3, 4))

    with pytest.raises(ValueError):
        FSCRotation3dMatrix((9, 24), data=data)

    with pytest.raises(ValueError):
        FSCRotation3dMatrix((10, 24), data=data)

    with pytest.raises(ValueError):
        FSCRotation3dMatrix((25, 9), data=data)

    with pytest.raises(TypeError):
        FSCRotation3dMatrix((24, 9), data=data_wrong1)

    with pytest.raises(TypeError):
        FSCRotation3dMatrix((24, 9), data=data_wrong2)

    with pytest.raises(TypeError):
        FSCRotation3dMatrix((24, 9), data=data_wrong3)

    with pytest.raises(ValueError):
        FSCRotation3dMatrix((24, 9), 5, data=data)

    with pytest.raises(ValueError):
        FSCRotation3dMatrix((24, 9))

    mat = FSCRotation3dMatrix((24, 9), data=data)
    with pytest.raises(ValueError):
        mat._matvec(np.ones(4))

    mat._matvec(np.ones((3, 3)))
    with pytest.raises(ValueError):
        mat._matvec(np.ones(25))

    with pytest.raises(ValueError):
        mat._matvec(np.ones((8, 3)))

    with pytest.raises(TypeError):
        mat._matvec(np.ones((3, 3)), out=1)

    with pytest.raises(ValueError):
        mat._matvec(np.ones((3, 3)), out=np.zeros(7))


def test_fsr_rot3d_error():
    data = np.zeros(
        (3, 4), dtype=[('index', int), ('r11', float), ('r22', float), ('r32', float)]
    )
    data_wrong1 = np.empty((3, 4), [('index', int), ('value', float)])
    data_wrong2 = np.empty(
        (3, 4), [('r11', float), ('r22', float), ('r32', float), ('index', int)]
    )
    data_wrong3 = np.empty(
        (3, 4),
        dtype=[('index', int), ('r11', float), ('r22', np.float32), ('r32', float)],
    )

    with pytest.raises(TypeError):
        FSRRotation3dMatrix(3)

    with pytest.raises(ValueError):
        FSRRotation3dMatrix((2, 3, 4))

    with pytest.raises(ValueError):
        FSRRotation3dMatrix((24, 9), data=data)

    with pytest.raises(ValueError):
        FSRRotation3dMatrix((10, 24), data=data)

    with pytest.raises(ValueError):
        FSRRotation3dMatrix((9, 25), data=data)

    with pytest.raises(TypeError):
        FSRRotation3dMatrix((9, 24), data=data_wrong1)

    with pytest.raises(TypeError):
        FSRRotation3dMatrix((9, 24), data=data_wrong2)

    with pytest.raises(TypeError):
        FSRRotation3dMatrix((9, 24), data=data_wrong3)

    with pytest.raises(ValueError):
        FSRRotation3dMatrix((9, 24), 5, data=data)

    with pytest.raises(ValueError):
        FSRRotation3dMatrix((9, 24))

    mat = FSRRotation3dMatrix((9, 24), data=data)
    with pytest.raises(ValueError):
        mat._matvec(np.ones(4))

    mat._matvec(np.ones((8, 3)))
    with pytest.raises(ValueError):
        mat._matvec(np.ones(25))

    with pytest.raises(ValueError):
        mat._matvec(np.ones((3, 3)))

    with pytest.raises(TypeError):
        mat._matvec(np.ones((8, 3)), out=1)

    with pytest.raises(ValueError):
        mat._matvec(np.ones((8, 3)), out=np.zeros(7))


def test_sparse_operator():
    with pytest.raises(TypeError):
        SparseOperator(np.zeros((10, 3)))
    sp = SparseOperator(scipy.sparse.csc_matrix((4, 6)))
    assert type(sp) is pyoperators.linear.SparseOperator

    data = FSRMatrix((10, 3), ncolmax=1)
    with pytest.raises(ValueError):
        SparseOperator(data, shapein=4)
    with pytest.raises(ValueError):
        SparseOperator(data, shapeout=11)

    data = FSRRotation2dMatrix((12, 6), ncolmax=1)
    with pytest.raises(ValueError):
        SparseOperator(data, shapein=(5, 2))
    with pytest.raises(ValueError):
        SparseOperator(data, shapeout=(7,))

    data = FSRRotation3dMatrix((12, 6), ncolmax=1)
    with pytest.raises(ValueError):
        SparseOperator(data, shapein=(5, 3))
    with pytest.raises(ValueError):
        SparseOperator(data, shapeout=(7,))



"""
This module provides utilities for the cartesian 2-d geometry.

The coordinates of the geometric object provided by this module refer
to a cartesian referential (X, Y). To state it explicitly, for a given array
of coordinates a, a[..., 0] refers to the X-axis and a[..., 1] to the Y-axis.
           Y ^
             |
             +--> X
Angles are counted counter-clockwise.

"""

import numpy as _np
import scipy.spatial as _spatial

from . import _flib


def convex_hull(points):
    """
    Return the convex hull of 2-dimensional points. The points in the convex
    hull are sorted counter-clockwise.

    Parameters
    ----------
    points : array-like of shape (..., 2)
        The 2-dimensional points for which the convex hull is computed.

    """
    points = _np.atleast_1d(points)
    if points.shape[-1] != 2:
        raise ValueError(
            f"Invalid last dimension of 2-dimensional points '{points.shape[-1]}'."
        )
    points = points.reshape((-1, 2))

    def arctan2(v):
        return (_np.arctan2(v[:, 1], v[:, 0]) + 2 * _np.pi) % (2 * _np.pi)

    ihull = list(set(_spatial.Delaunay(points).convex_hull.ravel()))
    hull = points[ihull, :]
    return hull[_np.argsort(arctan2(hull - hull.mean(axis=0))), :]


def create_circle(radius, out=None, center=(0, 0), n=72, dtype=float):
    """
    Return coordinates of a circle.

    Parameters
    ----------
    radius : float
        The circle radius.
    out : array of shape radius.shape + (2,), optional
        Placeholder for the output coordinates.
    center : array-like of size 2, optional
        The (X, Y) coordinates of the circle center.
    n : integer
        Number of vertices, optional
    dtype : dtype, optional
        Coordinate data type.

    """
    return create_regular_polygon(n, radius, out=out, center=center, dtype=dtype)


def create_grid(
    shape,
    spacing,
    out=None,
    xreflection=False,
    yreflection=False,
    center=(0, 0),
    angle=0,
):
    """
    The grid nodes are laid out as a rectangular matrix (nrows, ncolumns)
    in the (X, Y) referential. Before rotation, rows increase along the -Y axis
    (unless yreflection is set to True) and columns increase along the X axis
    (unless xreflection is set to True).

    Parameters
    ----------
    shape : tuple of two integers (nrows, ncolumns)
        Number of rows and columns of the grid.
    spacing : float, Quantity
        Physical spacing between grid nodes.
    out : array of shape (shape[0], shape[1], 2), optional
        Placeholder for the output coordinates.
    xreflection : boolean, optional
        Reflection along the X-axis before rotation and translation.
    yreflection : boolean, optional
        Reflection along the Y-axis before rotation and translation.
    center : array-like of shape (2,), optional
        The (X, Y) coordinates of the grid center
    angle : float, optional
        Counter-clockwise rotation in degrees around the grid center.

    """
    if out is None:
        out = _np.empty(shape + (2,))
    _flib.geometry.create_grid(
        spacing, xreflection, yreflection, angle, center[0], center[1], out.T
    )
    return out


def create_grid_squares(
    shape,
    spacing,
    out=None,
    filling_factor=1,
    xreflection=False,
    yreflection=False,
    center=(0, 0),
    angle=0,
):
    """
    The square centers are laid out as a rectangular matrix (nrows, ncolumns)
    in the (X, Y) referential. Before rotation, rows increase along the -Y axis
    (unless yreflection is set to True) and columns increase along the X axis
    (unless xreflection is set to True).

    Parameters
    ----------
    shape : tuple of two integers (nrows, ncolumns)
        Number of rows and columns of the grid.
    spacing : float, Quantity
        Physical spacing between square centers.
    out : array of shape (shape[0], shape[1], 4, 2), optional
        Placeholder for the output coordinates.
    filling_factor : float, optional
        Fraction of the area of the squares relative to spacing**2.
    xreflection : boolean, optional
        Reflection along the X-axis before rotation and translation.
    yreflection : boolean, optional
        Reflection along the Y-axis before rotation and translation.
    center : array-like of shape (2,), optional
        The (X, Y) coordinates of the grid center
    angle : float, optional
        Counter-clockwise rotation in degrees around the grid center.

    """
    if out is None:
        out = _np.empty(shape + (4, 2))
    _flib.geometry.create_grid_squares(
        spacing,
        filling_factor,
        xreflection,
        yreflection,
        angle,
        center[0],
        center[1],
        out.T,
    )
    return out


def create_rectangle(size, out=None, center=(0, 0), angle=0, dtype=float):
    """
    Return coordinates of a rectangle.

    Parameters
    ----------
    size : array-like of size 2
        The rectangle dimensions along the X and Y axis.
    out : array of shape size.shape[:-1] + (4, 2), optional
        Placeholder for the output coordinates.
    center : array-like of size 2, optional
        The (X, Y) coordinates of the rectangle center.
    angle : float, optional
        Angle of counter-clockwise rotation around the center in degrees.
    dtype : dtype, optional
        Coordinate data type.

    """
    size = _np.asarray(size, dtype=dtype)
    if size.shape[-1] != 2:
        raise ValueError('Invalid rectangle dimensions.')
    if out is None:
        out = _np.empty(size.shape[:-1] + (4, 2), dtype)
    out[...] = (
        _np.asarray([[0.5, 0.5], [-0.5, 0.5], [-0.5, -0.5], [0.5, -0.5]], dtype)
        * size[..., None, :]
    )
    if angle != 0:
        rotate(out, angle, out=out)
    out += center
    return out


def create_regular_polygon(n, radius, out=None, center=(0, 0), angle=0, dtype=float):
    """
    Return coordinates of a regular polygon.

    Parameters
    ----------
    n : integer
        Number of vertices, optional
    radius : float
        The center-to-vertex distance.
    out : array of shape radius.shape + (2,), optional
        Placeholder for the output coordinates.
    center : array-like of size 2, optional
        The (X, Y) coordinates of the circle center.
    angle : float, optional
        Counter-clockwise rotation in degrees around center.
    dtype : dtype, optional
        Coordinate data type.

    """
    radius = _np.asarray(radius, dtype)
    center = _np.asarray(center, dtype)
    if out is None:
        b = _np.broadcast(radius[..., None], center)
        out = _np.empty(b.shape[:-1] + (n, 2), dtype)
    pi = 4 * _np.arctan(_np.asarray(1, dtype))
    a = 2 * pi / n * _np.arange(n, dtype=dtype) + _np.radians(angle)
    out[...] = (
        radius[..., None, None] * _np.asarray([_np.cos(a), _np.sin(a)]).T
        + center[..., None, :]
    )
    return out


def create_square(size, out=None, center=(0, 0), angle=0, dtype=float):
    """
    Return coordinates of a square.

    Parameters
    ----------
    size : float
        The square size.
    out : array of shape size.shape + (4, 2), optional
        Placeholder for the output coordinates.
    center : array-like of size 2, optional
        The (X, Y) coordinates of the square center.
    angle : float, optional
        Angle of counter-clockwise rotation around the center in degrees.
    dtype : dtype, optional
        Coordinate data type.

    """
    size = _np.asarray(size, dtype=dtype)
    if out is None:
        out = _np.empty(size.shape + (4, 2), dtype)
    out[...] = (
        _np.asarray([[0.5, 0.5], [-0.5, 0.5], [-0.5, -0.5], [0.5, -0.5]], dtype)
        * size[..., None, None]
    )
    if angle != 0:
        rotate(out, angle, out=out)
    out += center
    return out


def segment_polygon(vertices, max_distance, dtype=float):
    """
    Segment the edges of a polygon, ensuring that all segments have a size
    lower than a given distance. Such segmentation can be useful to better
    represent a polygon after a spatial transform.

    Parameters
    ----------
    vertices : array-like, last dimension is 2
        The (X,Y) coordinates of the vertices.
    min_distance : float
        The maximum size of the segments.
    dtype : dtype, optional
        If provided, forces the calculation to use the data type specified.

    """
    vertices = _np.array(vertices, dtype)
    if vertices.ndim != 2:
        raise ValueError('Invalid dimension of the input vertices.')
    vertices_ = _np.vstack([vertices, vertices[0]])
    ds = [_np.hypot(c[0], c[1]) for c in _np.diff(vertices_, axis=0)]
    ns = [int(_np.ceil(d / max_distance)) for d in ds]
    coords = _np.empty((_np.sum(ns), 2), vertices.dtype)
    i = 0
    for v0, v1, n in zip(vertices_[:-1], vertices_[1:], ns):
        alpha = _np.arange(n) / n
        coords[i : i + n, :] = _np.outer(1 - alpha, v0) + _np.outer(alpha, v1)
        i += n
    return coords


def rotate(coords, angle, out=None):
    """
    Rotate one or more points around the origin.

    Parameters
    ----------
    coords : array-like, last dimension is 2
        The (X, Y) coordinates of the points.
    angle : float
        Angle of counter-clockwise rotation in degrees.
    out : ndarray, optional
        If provided, the calculation is done into this array.

    """
    from pyoperators.utils import isalias

    coords = _np.asarray(coords)
    if out is None:
        out = _np.empty_like(coords)
    if (
        coords.dtype.char == 'd'
        and coords.flags.c_contiguous
        and out.flags.c_contiguous
    ):
        if isalias(coords, out):
            _flib.geometry.rotate_2d_inplace(out.reshape((-1, 2)).T, angle)
        else:
            _flib.geometry.rotate_2d(
                coords.reshape((-1, 2)).T, out.reshape((-1, 2)).T, angle
            )
        return out
    if coords.dtype.itemsize > 8:
        angle = _np.asarray(angle, coords.dtype)
    angle = _np.radians(angle)
    m = _np.asarray(
        [[_np.cos(angle), -_np.sin(angle)], [_np.sin(angle), _np.cos(angle)]],
        coords.dtype,
    )
    if isalias(coords, out):
        coords = coords.copy()
    return _np.dot(coords, m.T, out)


def surface_simple_polygon(coords, out=None, dtype=float):
    """
    Return the surface of convex polygons.

    Parameters
    ----------
    coords : array-like of shape (..., nvertices, 2)
        The vertices coordinates.
    out : ndarray, optional
        Placeholder of the polygon surfaces.

    Example
    -------
    >>> from pysimulators.geometry import (
    ...     create_square, create_regular_polygon, surface_simple_polygon)
    >>> c = create_square([1, 2])
    >>> surface_simple_polygon(c)
    array([ 1.,  4.])
    >>> surface_simple_polygon(create_regular_polygon(6, 1))

    """
    coords = _np.array(coords, dtype=float, order='c')
    if out is None:
        out = _np.empty(coords.shape[:-2], dtype)
    if coords.ndim < 3:
        coords = coords[None, ...]
    if out.ndim == 0:
        out_ = out.reshape(1)
    else:
        out_ = out
    _flib.geometry.surface_simple_polygon(coords.T, out_.T)
    if out.ndim == 0:
        out = out[()]
    return out



# Copyrights 2010-2013 Pierre Chanial
# All rights reserved


import gc
import operator
from copy import copy

import numpy as np

from pyoperators import (
    MPI,
    BlockColumnOperator,
    BlockDiagonalOperator,
    CompositionOperator,
)
from pyoperators.memory import empty
from pyoperators.utils import isscalarlike, operation_assignment, split, strenum

from .instruments import Instrument
from .packedtables import Layout, Sampling, Scene

__all__ = ['Acquisition', 'MaskPolicy']


class Acquisition:
    """
    The Acquisition class, which combines the instrument, sampling and scene
    models.

    """

    def __init__(
        self,
        instrument,
        sampling,
        scene,
        block=None,
        max_nbytes=None,
        nprocs_instrument=None,
        nprocs_sampling=None,
        comm=None,
    ):
        """
        Parameters
        ----------
        instrument : Instrument
            The Instrument instance.
        sampling : Sampling
            The sampling information (pointings, etc.)
        scene : Scene
            Discretization of the observed scene.
        block : tuple of slices, optional
            Partition of the samplings.
        max_nbytes : int or None, optional
            Maximum number of bytes to be allocated for the acquisition's
            operator.
        nprocs_instrument : int
            For a given sampling slice, number of procs dedicated to
            the instrument.
        nprocs_sampling : int
            For a given detector slice, number of procs dedicated to
            the sampling.
        comm : mpi4py.MPI.Comm
            The acquisition's MPI communicator. Note that it is transformed
            into a 2d cartesian communicator before being stored as the 'comm'
            attribute. The following relationship must hold:
                comm.size = nprocs_instrument * nprocs_sampling

        """
        if not isinstance(instrument, Instrument):
            raise TypeError(
                f'The instrument input has an invalid type '
                f'{type(instrument).__name__!r}.'
            )
        if not isinstance(sampling, Sampling):
            raise TypeError(
                f'The sampling input has an invalid type {type(instrument).__name__!r}.'
            )
        if not isinstance(scene, Scene):
            raise TypeError(
                f'The scene input has an invalid type {type(scene).__name__!r}.'
            )

        if comm is None:
            comm = MPI.COMM_WORLD
        if nprocs_instrument is None and nprocs_sampling is None:
            nprocs_instrument = 1
            nprocs_sampling = comm.size
        elif nprocs_instrument is None:
            if nprocs_sampling < 1 or nprocs_sampling > comm.size:
                raise ValueError(
                    f"Invalid value for nprocs_sampling '{nprocs_sampling}'."
                )
            nprocs_instrument = comm.size // nprocs_sampling
        else:
            if nprocs_instrument < 1 or nprocs_instrument > comm.size:
                raise ValueError(
                    f"Invalid value for nprocs_instrument '{nprocs_instrument}'."
                )
            nprocs_sampling = comm.size // nprocs_instrument
        if nprocs_instrument * nprocs_sampling != comm.size:
            raise ValueError('Invalid MPI distribution of the acquisition.')

        commgrid = comm.Create_cart([nprocs_sampling, nprocs_instrument], reorder=True)

        comm_instrument = commgrid.Sub([False, True])
        comm_sampling = commgrid.Sub([True, False])

        self.scene = scene
        self.instrument = instrument.scatter(comm_instrument)
        self.sampling = sampling.scatter(comm_sampling)
        self.comm = commgrid
        self.block = block
        if block is None:
            self.block = (slice(0, len(self.sampling)),)
            if max_nbytes is not None:
                nbytes = self.get_operator_nbytes()
                if nbytes > max_nbytes:
                    nblocks = int(np.ceil(nbytes / max_nbytes))
                    self.block = tuple(split(len(self.sampling), nblocks))
        elif not isinstance(block, (list, tuple)) or any(
            not isinstance(b, slice) for b in block
        ):
            raise TypeError(f"Invalid block argument: '{block}'.")

    _operator = None

    def __getitem__(self, x):
        """
        Restrict the acquisition model to a set of detectors, samplings or
        scene pixels.

        new_acq = acq[selection_instrument, selection_sampling,
                      selection_scene]

        Example
        -------
        >>> acq = MyAcquisition()
        Restrict to the first 10 detectors:
        >>> new_acq = acq[:10]
        Restrict to the first 10 samplings:
        >>> new_acq = acq[:, :10]
        Restrict to the first 10 pixels of the scene:
        >>> new_acq = acq[..., :10]
        """

        def is_colon(x):
            return isinstance(x, slice) and x == slice(None)

        out = copy(self)
        if not isinstance(x, tuple):
            x = (x,)
        elif len(x) == 2 and x[0] is Ellipsis:
            x = slice(None), slice(None), x[1]
        elif len(x) > 3:
            raise ValueError('Invalid selection.')

        x = tuple(slice(None) if _ is Ellipsis else _ for _ in x)
        x = x + (3 - len(x)) * (slice(None),)

        if all(is_colon(_) for _ in x):
            return out

        if any(not is_colon(_) for _ in x):
            self._operator = None
            gc.collect()

        out.instrument = self.instrument[x[0]]
        if not is_colon(x[0]):
            object.__setattr__(out.instrument.detector, 'comm', MPI.COMM_SELF)
        out.sampling = self.sampling[x[1]]  # XXX FIX BLOCKS!!!
        if not is_colon(x[1]):
            object.__setattr__(out.sampling, 'comm', MPI.COMM_SELF)
        out.scene = self.scene[x[2]]

        if not is_colon(x[0]) and not is_colon(x[1]):
            out.comm = MPI.COMM_SELF
        elif not is_colon(x[0]):
            out.comm = self.sampling.comm.Create_cart([1, self.sampling.comm.size])
        elif not is_colon(x[1]):
            out.comm = self.instrument.comm.Create_cart([self.instrument.comm.size, 1])

        return out

    def __str__(self):
        return f'{self.instrument}\nSamplings: {len(self.sampling)}'

    __repr__ = __str__

    def pack(self, x, out=None, copy=False):
        return self.instrument.detector.pack(x)

    pack.__doc__ = Layout.pack.__doc__

    def unpack(self, x, out=None, missing_value=None, copy=False):
        return self.instrument.detector.unpack(x)

    unpack.__doc__ = Layout.unpack.__doc__

    def get_observation(
        self, x, noiseless=False, out=None, operation=operation_assignment
    ):
        """
        Return out=H(x)+n, the observation according to the acquisition model.

        Parameters
        ----------
        x : float array
            The input values of the Scene.
        noiseless : boolean, optional
            If True, no noise is added to the observation.
        out : float array
            Buffer of shape len(instrument)xlen(sampling) for the output
            observation.
        operation : function
            get_observation(x, out=out, operation=operation) is equivalent to
            operation(out, get_observation(x))

        """
        H = self.get_operator()
        out = H(x, out, operation=operation)
        if not noiseless:
            self.get_noise(out, operation=operator.iadd)
        return out

    def get_operator(self):
        """
        Return the acquisition model H as an operator.

        """
        if self._operator is None:
            self._operator = CompositionOperator(
                [
                    BlockColumnOperator(
                        [
                            self.instrument.get_operator(self.sampling[b], self.scene)
                            for b in self.block
                        ],
                        axisin=1,
                    ),
                    self.scene.get_distribution_operator(self.comm),
                ]
            )
        return self._operator

    def get_operator_nbytes(self):
        """
        Return the number of bytes required to store the acquisition model
        as an operator.

        """
        n = self[0, 0].get_operator().nbytes
        return n * len(self.instrument) * len(self.sampling)

    def get_invntt_operator(self):
        """
        Return the inverse time-time noise correlation matrix as an Operator.

        """
        return BlockDiagonalOperator(
            [self.instrument.get_invntt_operator(self.sampling[b]) for b in self.block],
            axisin=1,
        )

    def get_noise(self, out=None, operation=operation_assignment):
        """
        Return the noise realization according the instrument's noise model.

        Parameters
        ----------
        out : ndarray, optional
            Placeholder for the output noise.

        """
        if out is None:
            if operation is not operation_assignment:
                raise ValueError('The output buffer is not specified.')
            out = empty((len(self.instrument), len(self.sampling)))
        for b in self.block:
            self.instrument.get_noise(
                self.sampling[b], out=out[:, b], operation=operation
            )
        return out

    def plot(self, map=None, header=None, new_figure=True, percentile=0.01, **keywords):
        """
        map : ndarray of dim 2
            The optional map to be displayed as background.
        header : pyfits.Header
            The optional map's FITS header.
        new_figure : boolean
            If true, plot the scan in a new window.
        percentile : float, tuple of two floats
            As a float, percentile of values to be discarded, otherwise,
            percentile of the minimum and maximum values to be displayed.

        """
        if header is None:
            header = getattr(map, 'header', None)
            if header is None:
                header = self.sampling.get_map_header(naxis=1)
        annim = self.sampling.plot(
            map=map,
            header=header,
            new_figure=new_figure,
            percentile=percentile,
            **keywords,
        )
        return annim


class MaskPolicy:
    KEEP = 0
    MASK = 1
    REMOVE = 2

    def __init__(self, flags, values, description=None):
        self._description = description
        if isscalarlike(flags):
            if isinstance(flags, str):
                flags = flags.split(',')
            else:
                flags = [flags]
        flags = [_.strip() for _ in flags]
        if isscalarlike(values):
            if isinstance(values, str):
                values = values.split(',')
            else:
                values = [values]
        values = [_.strip().lower() for _ in values]
        if len(flags) != len(values):
            raise ValueError(
                'The number of policy flags is different from the'
                ' number of policies.'
            )

        self._policy = []
        for flag, value in zip(flags, values):
            if flag[0] == '_':
                raise ValueError(
                    'A policy flag should not start with an under' 'score.'
                )
            choices = 'keep', 'mask', 'remove'
            if value not in choices:
                raise KeyError(
                    f'Invalid policy {flag}={value!r}. Expected values are '
                    f'{strenum(choices)}.'
                )
            setattr(self, flag, value)
        self._flags = flags

    def __array__(self, dtype=int):
        conversion = {'keep': self.KEEP, 'mask': self.MASK, 'remove': self.REMOVE}
        return np.array(
            [conversion[getattr(self, _)] for _ in self._flags], dtype=dtype
        )

    def __str__(self):
        s = self._description + ': ' if self._description is not None else ''
        s += ', '.join(f'{_}={getattr(self, _)!r}' for _ in self._flags)
        return s
